from georunes.tools.filemanager import FileManager

filemanager = FileManager.get_instance()


class MassBalanceModalModeler:
    def __init__(self, data_minerals, ignore_oxides=None):
        raw_minerals_data = data_minerals
        self.ignore_oxels = ignore_oxides
        self.prepare_data(raw_minerals_data)
        self.modeler_name = "Mass balance modeler"
        self.init_name = 'Parental concentrations'

    def filter_elements(self, comp, verbose=1):
        removed = []
        for oxel in comp.keys():
            if oxel not in self.oxel_list:
                del comp[oxel]

        if len(removed) and verbose:
            print("Removed the oxels " + str(removed) + "from a composition.")

    def prepare_data(self, raw_minerals_data):

        if self.ignore_oxels:
            for ox in raw_minerals_data.columns.tolist()[1:]:
                if ox in self.ignore_oxels:
                    raw_minerals_data = raw_minerals_data.drop(columns=ox)
        raw_minerals_data = raw_minerals_data.fillna(0)

        # Remove row if all values are zeros
        raw_minerals_data = raw_minerals_data[(raw_minerals_data.iloc[:, 1:] != 0).any(axis=1)]
        self.raw_minerals_data = raw_minerals_data  # For optimization, index not needed
        raw_minerals_data = raw_minerals_data.set_index(raw_minerals_data.keys()[0])
        self.oxel_list = raw_minerals_data.columns.tolist()
        raw_minerals_data = raw_minerals_data[[*list(self.oxel_list)]]  # Order oxides as in source

        self.minerals_data = raw_minerals_data.transpose()
        self.default_minerals_data = raw_minerals_data.transpose()
        self.list_minerals = self.minerals_data.keys().tolist()
        self.nb_minerals = len(self.list_minerals)

    def get_elements_list(self):
        return self.minerals_data.index.tolist()

    def get_minerals_list(self):
        return self.minerals_data.columns

    def modal_props_to_concentration(self, modal_props):
        conc = dict()
        for oxel in self.oxel_list:
            sum_oxel = 0
            for mineral, fract in modal_props.items():
                sum_oxel += self.minerals_data[mineral][oxel] * fract
            conc[oxel] = sum_oxel / 100
        return conc

    def get_residual_liq_concentration(self, oxel, modal_cumulate, parent_conc, alpha_list):
        """
        Get the concentration of residual liquid for a list of alpha values (proportion of liquid)
        :param oxel: the oxide or the element (string)
        :param modal_cumulate: the modal mineral composition of the cumulate (dict)
        :param parent_conc: the concentrations of the parental melt (dict)
        :param alpha_list: the liquid proportions (list)
        :return: a list of the specified oxide/element concentration according to the liquid proportion
        """

        cumulate = self.modal_props_to_concentration(modal_cumulate)
        res = [(parent_conc[oxel] - alpha * cumulate[oxel]) / (1 - alpha) for alpha in alpha_list]
        return res

    def assimilate(self, parent, content, intake_ratio):
        """
        Calculate the assimilation of a content by a parental melt
        :param parent: the concentrations of the parental melt (dict)
        :param content: the concentrations of the assimilated component (dict)
        :param intake_ratio: the mass ratio of the assimilated content to the parental melt (float)
        :return: the concentrations of the mixed product (dict)
        """

    def set_minerals_data(self, minerals_data):
        self.minerals_data = minerals_data

    def reset_minerals_data(self):
        self.minerals_data = self.default_minerals_data.copy()

    @staticmethod
    def compute_alpha_limit_removal(parent, cumulate):
        div = parent / cumulate
        min_alpha = div.min()
        return min_alpha



import warnings
import numpy as np
from pandas import DataFrame
from georunes.modmin.optim.base import Optimizer, is_in_bounds, random_part_with_bounds
from georunes.tools.warnings import FunctionParameterWarning


class GradientDescent(Optimizer):

    def __init__(self, dist_func="euclidian", filling_tolerance=0.05, **kwargs):
        Optimizer.__init__(self, **kwargs)
        self.dist_func = dist_func
        self.filling_tolerance = filling_tolerance
        self.notif = ">>>>>> Gradient Descent method / deviation function : " + dist_func

    @staticmethod
    def grad(x, y, A, total=1):
        B = A.copy()
        B.loc[len(B.index)] = np.array([1] * len(x))  # Row added to minerals_data (A) to keep a sum equals to total
        Bt = B.transpose()
        y = np.append(y, total)
        out = (np.dot(np.dot(Bt, B), x) - np.dot(Bt, y)) / (len(B.index))
        return out

    def _compute(self, raw_data, skip_cols, raw_minerals_data, to_round=4,
                 max_minerals=None, min_minerals=None, unfillable_partitions_allowed=True, ignore_oxides=None,
                 max_iter=int(1e6), learn_rate=None, tolerance=1e-08, starting_partition=None, force_totals=False,
                 residual_in_suppl=False):

        if not learn_rate:
            learn_rate = 1 * pow(10, -to_round)

        if self.verbose > 1:
            print("Round digits :", to_round, " --  Maximum iterations :", max_iter, " --  Learn rate :", learn_rate,
                  " --  Tolerance :", tolerance,
                  " --  Oxides to ignore : " + str(ignore_oxides) if ignore_oxides else "")
            if starting_partition: print("Starting partition :", starting_partition)

        self.prepare_data(raw_data, skip_cols, raw_minerals_data, ignore_oxides)
        partitions = DataFrame(columns=self.list_minerals)
        deviation_name = "deviation_" + self.dist_func
        suppl = DataFrame(columns=[deviation_name])
        bulk_chems = [0]*len(self.data.index)
        found_chems = [0]*len(self.data.index)

        if force_totals:
            if unfillable_partitions_allowed:
                warnings.warn("The parameter force_totals is True. Then, the parameter unfillable_partitions_allowed "
                                "will be set to False.", FunctionParameterWarning)
            target_totals = self.init_total
        else:
            target_totals = [100] * len(self.data.index)

        # Get maximum proportion for each mineral
        for i in self.data.index:
            if self.verbose: print(">>> Composition", i)
            list_minerals_i = self.list_minerals.copy()
            minerals_data_i = self.minerals_data.copy()
            bulk_chems[i] = self.data.iloc[i].to_numpy()

            # Get maximum possible proportion for each mineral
            max_minerals_prop = []
            unnecessary_minerals = []
            for mineral in self.list_minerals:
                # Maximum proportion for each oxide equals Ox wt% in bulk chemistry / ox wt% in mineral composition
                max_oxides_prop = self.data.iloc[i] / minerals_data_i[mineral].transpose()
                mineral_prop = max_oxides_prop.min()
                if np.isinf(mineral_prop) or mineral_prop == 0:
                    minerals_data_i = minerals_data_i.drop(columns=mineral)
                    list_minerals_i.remove(mineral)
                    unnecessary_minerals.append(mineral)
                else:
                    max_minerals_prop.append(mineral_prop)
            nb_minerals_i = len(list_minerals_i)
            min_minerals_prop = [0] * nb_minerals_i

            # Update minerals proportions with defined settings if provided
            if max_minerals:
                for key, value in max_minerals.items():
                    idx = list_minerals_i.index(key)
                    if idx:
                        max_minerals_prop[idx] = min(max_minerals_prop[idx], max_minerals[key])
            if min_minerals:
                for key, value in min_minerals.items():
                    idx = list_minerals_i.index(key)
                    if idx:
                        min_minerals_prop[idx] = max(min_minerals_prop[idx], min_minerals[key])
            if not unfillable_partitions_allowed and sum(max_minerals_prop) < 1+self.filling_tolerance:  # % tolerance, default 5
                raise Exception("Problem in composition " + str(i)
                                + ". The sum of the maximum proportions of minerals cannot complete to 100 "
                                  "(found " + str(100 * sum(max_minerals_prop)) + str(")."))

            if self.verbose and unnecessary_minerals:
                print("Unnecessary minerals :", *unnecessary_minerals)

            # Gradient descent
            if self.verbose: print("Start calculations. Expected number of iterations :", max_iter)

            # Starting value
            if isinstance(starting_partition, dict):
                if not unfillable_partitions_allowed and sum(starting_partition.values()) != 100:
                    print("WARNING : The starting partition does not complete to 100. The calculations will be "
                          "started with a random composition.")
                    candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                        verbose=self.verbose)
                else:
                    # Check if all minerals in the config are present in minerals_data_i
                    if all(el in list_minerals_i for el in starting_partition.keys()):
                        candidate = []
                        for key in list_minerals_i:
                            val = starting_partition[key] / 100. if key in starting_partition.keys() else 0
                            candidate.append(val)
                        if not is_in_bounds(candidate, max_minerals_prop, min_minerals_prop):
                            print("WARNING : The starting partition does not respect the provided or the calculated "
                                  "bounds data : Minerals =", list_minerals_i, ", Max =", max_minerals_prop, ", Min =",
                                  min_minerals_prop, "The calculations will be started with a random composition.")
                            candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                                unfillable_partitions_allowed=unfillable_partitions_allowed, verbose=self.verbose)
                    else:
                        print("WARNING : Some minerals in starting partition are not present in mineral chemistry "
                              "data. The calculations will be started with a random composition.")
                        candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                            unfillable_partitions_allowed=unfillable_partitions_allowed,
                                                            verbose=self.verbose)
            else:
                candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                    unfillable_partitions_allowed=unfillable_partitions_allowed,
                                                    verbose=self.verbose)

            if self.verbose: print("Starting partition", dict(zip(list_minerals_i, candidate)))

            # Loop
            for k in range(max_iter):
                diff = -learn_rate * np.array(
                    self.grad(candidate, bulk_chems[i], minerals_data_i, total=target_totals[i] / 100))
                new_candidate = candidate + diff
                new_candidate = np.where(new_candidate < min_minerals_prop, min_minerals_prop, new_candidate)
                new_candidate = np.where(new_candidate > max_minerals_prop, max_minerals_prop, new_candidate)
                clean_diff = new_candidate - candidate

                if np.all(np.abs(clean_diff) <= tolerance):
                    if self.verbose: print("Tolerance condition reached after iteration", k)
                    break

                candidate = new_candidate
                if self.verbose > 2:
                    print("New solution at iteration", k)
                    print(dict(zip(list_minerals_i, candidate)))
                    corresp_chem = np.dot(minerals_data_i, candidate).round(decimals=to_round)
                    print("New deviation :", self.deviation(bulk_chems[i], corresp_chem),
                          "%" if self.dist_func == "SMAPE" else "")

            dict_res = {list_minerals_i[i]: round(candidate[i], to_round) for i in range(len(list_minerals_i))}

            idx_new_row = len(partitions)
            for miner, prop in dict_res.items():
                partitions.loc[idx_new_row, miner] = prop
            partitions = partitions.fillna(0).infer_objects(copy=False) # infer_objects prevent the downcasting of object dtype
            partitions.iloc[idx_new_row] = 100 * partitions.iloc[idx_new_row]
            partitions = partitions.round(to_round)

            deviation = self.deviation(bulk_chems[i], np.dot(self.minerals_data, partitions.iloc[idx_new_row] / 100.))
            found_chems[i] = np.dot(self.minerals_data, partitions.loc[idx_new_row] / 100.).round(to_round)
            suppl.loc[idx_new_row, deviation_name] = deviation
            suppl.loc[idx_new_row, "total_chem"] = sum(found_chems[i])

            if self.verbose:
                print("Solution", idx_new_row)
                print(partitions.loc[idx_new_row].to_dict())
                print("Corresponding composition")
                print([ str(a) + " : " + str(b) for a,b in zip(self.list_bulk_ox , found_chems[i])])
                print("Deviation :", round(deviation, to_round), "%" if self.dist_func == "SMAPE" else "", "\n------")

        partitions["Total"] = partitions.sum(axis=1).round(to_round)
        suppl["diff_total_chem"] = suppl["total_chem"] - self.init_total
        if residual_in_suppl:
            residuals = np.subtract(bulk_chems, found_chems)
            t_residuals = list(map(list, zip(*residuals)))  # Transposition
            for p, ox in enumerate(self.list_bulk_ox):
                print(t_residuals[p])
                suppl["resid_" + str(ox)] = t_residuals[p]

        return partitions, suppl



import tkinter as tk
from abc import abstractmethod, ABC
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure

_fract_values = [0., 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]


def check_float_list_and_order(lst, order="up"):

    # Validate input type and float range
    if isinstance(lst, list):
        for x in lst:
            if not isinstance(x, float) or not (0 <= x <= 1):
                return False

        # Check ordering function
        def is_sorted(l, ascending=True):
            for i in range(len(l) - 1):
                if ascending and l[i] > l[i + 1]:
                    return False
                if not ascending and l[i] < l[i + 1]:
                    return False
            return True

        ascending = (order == "up")

        if is_sorted(lst, ascending=ascending):
            return lst
        else:
            return sorted(lst, reverse=not ascending)
    else :
        return _fract_values

def total_label_color(total):
    if total == 100:
        total_lbl = "100"
        total_lbl_color = "#080"
    else:
        total_lbl = str(total)
        total_lbl_color = "#f00"
    return total_lbl, total_lbl_color


class BaseViewerTk(ABC):
    def __init__(self, petro_model, suppl_wm_title=None, liq_fract_values=None, verbose=0):
        self.verbose = verbose
        self.model = petro_model
        self.liq_fract_values = check_float_list_and_order(liq_fract_values,)

        # Window
        self.fig = Figure(figsize=(6, 4), dpi=100)
        self.ax = self.fig.add_subplot()
        self.root = tk.Tk()
        wm_title = "GeoRunes - PetroMod" + " - " + self.model.modeler_name

        if suppl_wm_title:
            wm_title = wm_title + " - " + suppl_wm_title
        self.root.wm_title(wm_title)
        self.show_legend = True

        # Frame
        self.frame = tk.Toplevel(self.root, )
        self.frame.minsize(300, 100)  # Minimum width
        self.plot_title = None

        # matplotlib canvas
        self.set_mpl_canvas()

    def set_supplement_status(self):
        pass

    def adjust_frame_position(self):
        # Frame position and settings
        self.root.update()
        dx = self.root.winfo_x() + self.root.winfo_width()
        dy = self.root.winfo_y()
        self.frame.geometry("+%d+%d" % (dx, dy))
        self.frame.protocol("WM_DELETE_WINDOW", self.on_closing_frame)

    def set_mpl_canvas(self):
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.root)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH)

        # Toolbar
        self.toolbar = NavigationToolbar2Tk(self.canvas, self.root, pack_toolbar=False)
        self.toolbar.update()
        self.toolbar.pack(side=tk.BOTTOM, fill=tk.X)

    def fill_frame(self):
        # Structure
        self.toolbar = tk.Frame(self.frame, height=40)
        self.headerbar = tk.Frame()
        self.content = tk.PanedWindow(self.frame)
        self.bottombar = tk.Frame(self.frame, height=60)
        self.statusbar = tk.Frame(self.frame, height=30)

        self.toolbar.pack(side="top", fill="x")
        self.statusbar.pack(side="bottom", fill="x")
        self.bottombar.pack(side="bottom", fill="x")
        self.content.pack(side="top", fill="both", expand=True)

    @abstractmethod
    def draw(self, element=None):
        pass

    def toggle_legend(self):
        self.show_legend = not self.show_legend
        self.refresh_plotting()

    def loop(self):
        self.root.mainloop()

    def refresh_plotting(self):
        self.fig.clear()
        self.ax = self.fig.add_subplot()
        self.draw()

    def on_closing_frame(self):
        self.canvas.get_tk_widget().destroy()
        self.root.destroy()

    def extra_widgets(self):
        pass



import random
import numpy as np
from matplotlib import colors as mplcol
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap


def darken_color(color, percent=10):
    fcol = list(mplcol.to_rgba(color))
    for i in range(4):
        fcol[i] = max(0, fcol[i] * (1 - percent / 100))
    return fcol


def near_color(color):
    col = mplcol.to_rgba_array(color)[0]

    cr = col[0]
    cg = col[1]
    cb = col[2]

    cr = cr + (-100 + random.randint(0, 200)) / 255.
    cg = cg + (-100 + random.randint(0, 200)) / 255.
    cb = cb + (-100 + random.randint(0, 200)) / 255.

    return (cr, cg, cb, 1)


def merge_colormaps(cmap1_name, cmap2_name, ratio=0.5, name='merged_cmap'):
    """
    Merge two matplotlib colormaps into one.

    Parameters:
        cmap1_name (str): Name of the first colormap (e.g., 'Blues')
        cmap2_name (str): Name of the second colormap (e.g., 'Oranges')
        ratio (float): Fraction of the output colormap to use from cmap1 (0 < ratio < 1)
        name (str): Name of the new colormap

    Returns:
        LinearSegmentedColormap: Merged colormap
    """
    assert 0 < ratio < 1, "Ratio must be between 0 and 1"

    n = 256
    n1 = int(n * ratio)
    n2 = n - n1

    cmap1 = plt.get_cmap(cmap1_name)
    cmap2 = plt.get_cmap(cmap2_name)

    colors1 = cmap1(np.linspace(0, 1, n1))
    colors2 = cmap2(np.linspace(0, 1, n2))
    combined = np.vstack((colors1, colors2))

    return LinearSegmentedColormap.from_list(name, combined)


import math

import matplotlib.pyplot as plt
import ternary
from matplotlib.colors import to_rgba

from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer
from georunes.tools.plotting import normalize_marker_size


class DiagramTernaryBase(DiagramBase, LegendDrawer):
    def __init__(self, datasource,
                 top_var='', left_var='', right_var='',
                 top_label='', left_label='', right_label='',
                 tscale='linear', left_scale='linear', right_scale='linear',
                 padding=None,
                 marker='', annotation=None,
                 alpha_color=0.4, alpha_edge_color=0.8,
                 legend_ncol=4,
                 no_ticks=False, no_ticks_label=False,
                 h_ratio=None, vertical_ticks=False,
                 scale=100,
                 **kwargs
                 ):
        self.scale = scale
        if padding is None:
            padding = {"bottom": 0.14}
        self.fig = None
        self.ax = None
        self.tax = None
        DiagramBase.__init__(self, datasource=datasource, h_ratio=h_ratio, legend_ncol=legend_ncol, padding=padding,
                             **kwargs)
        self.annotation = annotation
        self.top_label = top_label
        self.left_label = left_label
        self.right_label = right_label
        self.top_var = top_var
        self.left_var = left_var
        self.right_var = right_var
        self.top_scale = tscale
        self.left_scale = left_scale
        self.right_scale = right_scale
        self.marker = marker
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color
        self.no_ticks = no_ticks
        self.no_ticks_label = no_ticks_label
        self.vertical_ticks = vertical_ticks

    def init_plot(self):
        ratio_parameter = math.sqrt(3) / 2
        if self.h_ratio:
            ratio_parameter = self.h_ratio
        self.fig, self.ax = plt.subplots(figsize=plt.figaspect(ratio_parameter))  # Ratio figure size
        self.tax = ternary.TernaryAxesSubplot(ax=self.ax, scale=self.scale)

    def plot_config(self, no_gridline=False, no_ticks=False, no_ticks_label=False):
        DiagramBase.plot_config(self)

        if not no_gridline:
            self.tax.gridlines(color="black", multiple=10, linewidth=0.5)
        self.tax.boundary(linewidth=2.0)

        self.tax.left_corner_label(self.left_label)
        self.tax.right_corner_label(self.right_label)
        self.tax.top_corner_label(self.top_label)

        if not self.no_title:
            if self.title == "":
                self.title = 'Triangular diagram'
            self.tax.set_title(self.title)

        lw = 0 if no_ticks else 1

        if not no_ticks:
            if no_ticks_label:
                self.tax.ticks(multiple=10, axis='lbr', linewidth=lw, linestyle='-', tick_formats="")
            else:
                self.tax.ticks(multiple=10, axis='lbr', linewidth=lw, linestyle='-', fontsize=10, offset=0.02)

        if self.vertical_ticks:
            self.ax.set_xlim()
            self.ax.spines['bottom'].set_visible(False)
            self.ax.spines['top'].set_visible(False)
            self.ax.spines['right'].set_visible(False)
            self.ax.get_xaxis().set_visible(False)
        else:
            self.tax.get_axes().axis('off')

    def plot(self):
        self.plot_config(no_gridline=False, no_ticks=self.no_ticks, no_ticks_label=self.no_ticks_label)

        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):
                var_sum = group[self.top_var] + group[self.left_var] + group[self.right_var]
                norm_z = 100 * group[self.left_var] / var_sum
                norm_x = 100 * group[self.right_var] / var_sum
                norm_y = 100 * group[self.top_var] / var_sum

                if self.marker != '':
                    marker = self.marker
                else:
                    marker = list(group[self.marker_column])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = None
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                # Convert points to tuples
                points = [(norm_x.get(i), norm_y.get(i), norm_z.get(i)) for i, sample in
                          group[self.top_var].items()]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.tax.scatter(points, edgecolors=edge_color,
                                 marker=marker, label=label, facecolors=sample_color,
                                 s=size,
                                 alpha=self.alpha_color, zorder=zorder)

        self.plot_legend()
        self.tax.clear_matplotlib_ticks()



from matplotlib import pyplot
from georunes.plot.piper import helpers as pip
from georunes.plot.piper.helpers import split_by_field
from georunes.tools.data import repeat_list


def scatter(points, delta, ax=None, edgecolors=None, facecolors=None, **kwargs):
    """
    Plots points of composition (Ca, Mg, Na + K, HCO3+ CO3, SO4, Cl) in a Piper diagram

    Parameters
    ----------
    points: List of 3-tuples
        The list of tuples to be scatter-plotted.
    ax: PiperAxesSubplot, None
        The subplot to draw on.
    edgecolors :
        Colors of the marker's edges
    facecolors :
        Colors for filling the markers
    delta:
        The space separating the triangles (considering that the width of the diagram is scaled to 1).
    kwargs:
        Any kwargs to pass through to matplotlib.
    """
    if not ax:
        fig, ax = pyplot.subplots()
    xs, ys = pip.project_sequence(points, delta=delta)
    ax.scatter(xs, ys, edgecolors=repeat_list(edgecolors, len(xs)), facecolors=repeat_list(facecolors,  len(xs)),  **kwargs)
    return ax


def plot(points, delta, ax=None, **kwargs):
    """
    Plots trajectory points of composition (Ca, Mg, Na + K, HCO3+ CO3, SO4, Cl) in a Piper diagram

    Parameters
    ----------
    points: List of 3-tuples
        The list of tuples to be plotted as a connected curve.
    delta:
    ax: PiperAxesSubplot, None
        The subplot to draw on.
    kwargs:
        Any kwargs to pass through to matplotlib.
    """
    if not ax:
        fig, ax = pyplot.subplots()
    xs, ys = pip.project_sequence(points, delta)
    splitted = split_by_field(xs, ys)
    for xsp, ysp in splitted:
        ax.plot(xsp, ysp, **kwargs)
    return ax



import tkinter as tk
from tkinter import ttk
from georunes.petromod.modelers.partition import compute_bulk_coeffs
from georunes.tools.gui.tkwidets.entry_with_callback import EntryWithCallback
from georunes.tools.gui.tkwidets.editable_treeview import EditableFloatTreeview
from georunes.petromod.viewer.tk.base_viewer import BaseViewerTk, total_label_color
from georunes.tools.data import get_key, float_or_zero


class FractViewerTk(BaseViewerTk):
    def __init__(self, petro_model, data_coeffs, initial_conc, solid_comp=None, suppl_wm_title=None, *args, **kwargs):
        self.default_solid = solid_comp
        self.default_coeffs = data_coeffs
        self.default_initial_conc = initial_conc.copy()
        self.current_coeffs = self.default_coeffs.copy()
        self.coeffs_minerals = self.default_coeffs.keys()
        self.coeffs_elements = self.default_coeffs.index.tolist()
        self.initial_conc = initial_conc.copy()
        self.initial_conc_elts = initial_conc.index.tolist()
        self.set_supplement_status()
        self.init_solid_comp()

        BaseViewerTk.__init__(self, petro_model, suppl_wm_title=suppl_wm_title, *args, **kwargs)
        self.init_bulk_coeffs()

        # Concentration choice
        self.phases_labels = self.model.get_phases_labels()
        self.selected_phase = list(self.phases_labels.keys())[0]
        self.plot_title = list(self.phases_labels.values())[0]

        self.fill_frame()
        self.adjust_frame_position()

    def set_supplement_status(self):
        self.supplement_status = []

    def update_initial_conc(self):
        new_conc0 = self.initial_conc.copy()

        for label, entry in self.initial_conc_widgets:
            new_conc0[label.cget("text")] = float(entry.get())
        self.initial_conc = new_conc0

    def init_solid_comp(self):
        self.solid_comp = self.default_solid
        total = 0
        for element, val in self.default_solid.items():
            total += float_or_zero(val)
        total = round(total, 6)
        self.solid_total = total

        self.missing_minerals = []
        for mineral in self.solid_comp.keys():
            if mineral not in self.coeffs_minerals:
                self.missing_minerals.append(mineral)
        self.missing_elements = []
        for element in self.initial_conc_elts:
            if element not in self.coeffs_elements:
                self.missing_elements.append(element)

    def update_solid_comp(self):
        new_mineral_props = {}
        total = 0
        for label, entry in self.solid_comp_widgets:
            new_mineral_props[label.cget("text")] = float_or_zero(entry.get())
            total += float_or_zero(entry.get())
        total = round(total, 6)
        total_new_label, total_color = total_label_color(total)
        self.label_total.config(text=total_new_label, fg=total_color)
        self.solid_comp = new_mineral_props
        self.solid_total = total

    def get_selected_phase(self):
        return self.phase_selector.get()

    def update_selected_phase(self, _event):
        selected_label = self.get_selected_phase()
        selected_value = get_key(self.phases_labels, selected_label)
        self.selected_phase = selected_value
        self.plot_title = selected_label
        self.refresh_plotting()

    def refresh_model(self):
        self.update_initial_conc()
        self.update_solid_comp()
        self.update_bulk_coeffs()

    def refresh_model_and_plot(self):
        self.refresh_model()
        self.refresh_plotting()

    def update_coeffs(self, event):
        new_coeffs = self.coeffs_tree.data_tab
        self.current_coeffs = new_coeffs
        self.update_bulk_coeffs()
        self.refresh_bulk_coeffs_treeview()

    def _reset_solid(self):
        for index, (label, entry) in enumerate(self.solid_comp_widgets):
            entry.delete(0, tk.END)  # Clear the current content
            if label.cget("text") in self.default_solid.keys():
                entry.insert(0, self.default_solid[label.cget("text")])

    def event_reset_solid(self):
        self._reset_solid()

    def _reset_init_conc(self):
        for index, (label, entry) in enumerate(self.initial_conc_widgets):
            entry.delete(0, tk.END)  # Clear the current content
            entry.insert(0, self.default_initial_conc[label.cget("text")])  # Insert the new content

    def reset_init_conc(self):
        self._reset_init_conc()
        self.refresh_model()
        self.refresh_plotting()

    def _reset_coeffs(self):
        self.current_coeffs = self.default_coeffs.copy()
        self.update_bulk_coeffs()
        self.refresh_coeffs_treeview()

    def event_reset_coeffs(self):
        self._reset_coeffs()
        self.refresh_plotting()

    def refresh_coeffs_treeview(self):
        for row in self.coeffs_tree.get_children():
            self.coeffs_tree.delete(row)
        for index, row in self.current_coeffs.iterrows():
            self.coeffs_tree.insert("", tk.END, iid=index, values=[index] + row.tolist())

    def refresh_bulk_coeffs_treeview(self):
        for row in self.bulk_coeffs_tree.get_children():
            self.bulk_coeffs_tree.delete(row)
        for element, coeff in self.current_bulk_coeffs.items():
            self.bulk_coeffs_tree.insert("", tk.END, values=(element, coeff))

    def init_bulk_coeffs(self):
        new_bulk_coeffs = compute_bulk_coeffs(self.solid_comp, self.current_coeffs)
        self.model.set_bulk_dist_coeffs(new_bulk_coeffs)
        self.current_bulk_coeffs = new_bulk_coeffs
        if self.verbose > 0:
            print("New bulk partition coefficients")
            print(self.current_bulk_coeffs.to_dict())

    def update_bulk_coeffs(self):
        self.init_bulk_coeffs()
        self.refresh_bulk_coeffs_treeview()

    def draw(self, element=None):
        list_elements = self.model.list_elements()
        concentration_func = self.model.get_phase_concentration_func(self.selected_phase)
        self.ax.set_yscale('log')
        liq_vals = self.liq_fract_values
        liq_vals.sort()

        for element in list_elements:
            values = [concentration_func(element, f, self.initial_conc) for f in liq_vals]
            self.ax.plot(liq_vals, values, label=element, linestyle='-')
        self.ax.set_xlim((0, 1))
        self.ax.set_xlabel('Liquid fraction (F)')
        self.ax.set_ylabel('Concentration')
        self.ax.set_title(self.plot_title)
        self.ax.grid(axis='y', which='both', linestyle='-', color='#e8e8e8')
        if self.model.f_reverse is True:
            self.ax.invert_xaxis()
        if self.show_legend:
            self.ax.legend()
        self.canvas.draw()

    def fill_frame(self):
        BaseViewerTk.fill_frame(self)

        # Sideview
        self.sideview = tk.Frame(self.content)
        self.sideview.pack(side="left")
        pane_high = tk.Frame(self.sideview, )
        pane_high.pack(side='top')
        pane_low = tk.Frame(self.sideview, )
        pane_low.pack(side='bottom')
        pane_one = tk.Frame(self.sideview, width=140, padx=2, pady=2)
        pane_two = tk.Frame(self.sideview, width=140, padx=2, pady=2)
        pane_three = tk.Frame(self.sideview, width=140, padx=2, pady=2)
        pane_one.pack(side="left", anchor="n")
        pane_two.pack(side="left", anchor="n")
        pane_three.pack(side="left", anchor="n")

        pane_tab = tk.Frame(self.content, width=380, padx=2, pady=2)
        pane_tab.pack_propagate(False)
        pane_tab.pack(side="left", fill=tk.BOTH, anchor="n")

        # in sideview

        # Compositions
        comp_title = tk.Label(pane_one, text="Modal solid composition", anchor=tk.CENTER)
        comp_title.pack(side="top", )
        ttk.Separator(pane_one, orient='horizontal').pack(fill="x")

        proportions_pane = tk.Frame(pane_one)
        proportions_pane.pack(side="top")
        self.solid_comp_widgets = []
        total = 0
        for index, (oxel, value) in enumerate(self.solid_comp.items()):
            label = tk.Label(proportions_pane, text=oxel)
            entry = EntryWithCallback(proportions_pane, width=10, callback=self.refresh_model)
            entry.insert(0, value)
            entry.enable_callback()  # Enabled after initialization
            self.solid_comp_widgets.append((label, entry))
            total += value

        # Add other minerals existing in dataset
        for mineral in self.coeffs_minerals:
            if mineral not in self.solid_comp.keys():
                label = tk.Label(proportions_pane, text=mineral)
                entry = EntryWithCallback(proportions_pane, width=10, callback=self.refresh_model)
                entry.insert(0, '0')
                entry.enable_callback()  # Enabled after initialization
                self.solid_comp_widgets.append((label, entry))

        for index, (label, entry) in enumerate(self.solid_comp_widgets):
            label.grid(row=index, column=0, sticky=tk.W, pady=2)
            entry.grid(row=index, column=1, sticky=tk.W, pady=2)

        # Reset modal composition
        reset_solid_btn = tk.Button(pane_one, text="Reset solid composition", command=self.event_reset_solid)
        reset_solid_btn.pack()

        # Pane two
        tk.Label(pane_two, text=self.model.init_name, anchor=tk.CENTER).pack(side="top")
        ttk.Separator(pane_two, orient='horizontal').pack(side="top", fill="x")

        concentration_pane = tk.Frame(pane_two)
        concentration_pane.pack(side="top")
        self.initial_conc_widgets = []
        for index, (element, value) in enumerate(self.initial_conc.items()):
            label = tk.Label(concentration_pane, text=element)
            entry = tk.Entry(concentration_pane, width=10)
            entry.insert(0, value)
            self.initial_conc_widgets.append((label, entry))
        for index, (label, entry) in enumerate(self.initial_conc_widgets):
            label.grid(row=index, column=0, sticky=tk.W, pady=2)
            entry.grid(row=index, column=1, sticky=tk.W, pady=2)

        reset_conc_init_btn = tk.Button(pane_two, text="Reset initial composition", command=self.reset_init_conc)
        reset_conc_init_btn.pack()

        # Pane three
        bulk_coeffs_title = tk.Label(pane_three, text="Bulk partition coefficients", anchor=tk.CENTER)
        bulk_coeffs_title.pack(side="top", )

        # Create a Treeview widget to display the DataFrame
        self.bulk_coeffs_tree = ttk.Treeview(pane_three, columns=("col1", "col2"),
                                             show="headings", height=len(self.coeffs_elements))
        self.bulk_coeffs_tree.heading("col1", text="Element")
        self.bulk_coeffs_tree.heading("col2", text="Bulk distribution coeff.")
        self.bulk_coeffs_tree.column("col1", width=60, anchor="center")
        self.bulk_coeffs_tree.column("col2", width=60, anchor="center")
        for element, coeff in self.current_bulk_coeffs.items():
            self.bulk_coeffs_tree.insert("", "end", values=(element, coeff))

        self.bulk_coeffs_tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.bulk_coeffs_tree.bind('<<CellUpdated>>', self.update_coeffs)

        # Bottom bar
        ttk.Separator(self.bottombar, orient='horizontal').pack(side="top", fill="x")

        # Phase selection
        label_phase = tk.Label(self.bottombar, text="Phase selection", anchor=tk.CENTER)
        label_phase.pack(side="left", padx=5)
        self.phase_selector = ttk.Combobox(self.bottombar, width=30, values=list(self.phases_labels.values()))
        self.phase_selector.set(self.phases_labels[self.selected_phase])
        self.phase_selector.pack(side="left", padx=5)
        self.phase_selector.bind("<<ComboboxSelected>>", self.update_selected_phase)
        ttk.Separator(self.bottombar, orient='vertical').pack(side="left", padx=8, fill="y")

        check_legend = tk.Checkbutton(self.bottombar, text='Show/hide legend', command=self.toggle_legend)
        check_legend.select()
        check_legend.pack(side="left")

        # Refresh plotting
        ttk.Separator(self.bottombar, orient='vertical').pack(side="left", padx=8, fill="y")
        refresh_plotting_btn = tk.Button(self.bottombar, text="Refresh plotting", command=self.refresh_model_and_plot)
        refresh_plotting_btn.pack(side="left")

        # Status bar
        self.fill_status_bar()

        # Tab view
        # Coefficients Treeview

        label_coeffs_tree = tk.Label(pane_tab, text="Coefficients by mineral", anchor=tk.CENTER)
        label_coeffs_tree.pack(side="top")

        # Reset coefficients button in the bottom
        reset_coeffs_btn = tk.Button(pane_tab, text="Reset coefficients", command=self.event_reset_coeffs)
        reset_coeffs_btn.pack(side="bottom", )

        # Treeview
        self.coeffs_tree = EditableFloatTreeview(pane_tab, self.default_coeffs, )
        self.coeffs_tree.column('#0', width=0, stretch=tk.NO)  # Hide the first column

        # Define number of columns
        self.coeffs_tree['columns'] = ['Element'] + list(self.coeffs_minerals)

        # Format columns
        for col in self.current_coeffs.columns:
            self.coeffs_tree.column(col, width=60, stretch=False, anchor='w')  # width=100, minwidth=100
            self.coeffs_tree.heading(col, text=col)

        # Insert data into the Treeview
        for index, row in self.current_coeffs.iterrows():
            self.coeffs_tree.insert("", tk.END, iid=index, values=[index] + row.tolist())
        self.coeffs_tree.column('#1', width=40, stretch=False, anchor='w')

        # Create a horizontal scrollbar
        hscroll = ttk.Scrollbar(pane_tab, orient="horizontal", command=self.coeffs_tree.xview)
        hscroll.pack(side="bottom", fill=tk.X)

        self.coeffs_tree.configure(xscrollcommand=hscroll.set)
        self.coeffs_tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.coeffs_tree.bind('<<CellUpdated>>', self.update_coeffs)

        # Extra widgets
        self.extra_widgets()

    def fill_status_bar(self):
        ttk.Separator(self.statusbar, orient='horizontal').pack(side="top", fill="x")

        # Total
        total_label, total_color = total_label_color(self.solid_total)
        tk.Label(self.statusbar, text="Total :").pack(side="left")
        self.label_total = tk.Label(self.statusbar, text=total_label, fg=total_color)
        self.label_total.pack(side="left")

        # About missing minerals or missing oxels
        if len(self.missing_minerals) > 0:
            ttk.Separator(self.statusbar, orient='vertical').pack(side="left", fill="y")

            label_content = "Minerals missing in the dataset: " + str(self.missing_minerals)
            tk.Label(self.statusbar, text=label_content, fg="#f00").pack(side="left")

        if len(self.missing_elements) > 0:
            ttk.Separator(self.statusbar, orient='vertical').pack(side="left", fill="y")
            label_content2 = "Elements missing in the dataset: " + str(self.missing_elements)
            tk.Label(self.statusbar, text=label_content2, fg="#f00").pack(side="left")

        for content in self.supplement_status:
            ttk.Separator(self.statusbar, orient='vertical').pack(side="left", fill="y")
            tk.Label(self.statusbar, text=content, ).pack(side="left")



import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba
from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer, ArrowDrawer
from georunes.tools.chemistry import val_el_to_mol
from georunes.tools.plotting import normalize_marker_size


class DiagramScatter3D(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource,
                 xvar, yvar, zvar, xlim=None, ylim=None, zlim=None,
                 xlabel='', ylabel='', zlabel='',
                 annotation=None,
                 xscale='linear', yscale='linear', zscale='linear',
                 padding=None,
                 marker='',
                 xmolar=False, ymolar=False, zmolar=False,
                 alpha_color=0.4, alpha_edge_color=0.8,
                 x_formatter=None, y_formatter=None, z_formatter=None,
                 markersize=70,
                 **kwargs
                 ):
        config_padding = {"bottom": 0.20}
        if padding:
            config_padding.update(padding)
        self.xlabel = xlabel if xlabel else str(xvar)
        self.ylabel = ylabel if ylabel else str(yvar)
        self.zlabel = zlabel if zlabel else str(zvar)
        DiagramBase.__init__(self, datasource=datasource,
                             title=self.xlabel + " vs " + self.ylabel + " vs " + self.zlabel,
                             padding=config_padding,
                             markersize=markersize,
                             **kwargs)

        self.xlim = xlim
        self.ylim = ylim
        self.zlim = zlim
        self.xvar = xvar
        self.yvar = yvar
        self.zvar = zvar
        self.annotation = annotation
        self.xscale = xscale
        self.yscale = yscale
        self.zscale = zscale
        self.marker = marker
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color
        self.xmolar = xmolar
        self.ymolar = ymolar
        self.zmolar = zmolar
        self.x_formatter = x_formatter
        self.y_formatter = y_formatter
        self.z_formatter = z_formatter

    def init_plot(self):
        if self.h_ratio is None:
            self.fig = plt.figure()
            self.ax = self.fig.add_subplot(111, projection='3d')
        else:
            if isinstance(self.h_ratio, tuple):
                self.fig = plt.figure(figsize=self.h_ratio)
                self.ax = self.fig.add_subplot(111, projection='3d')
            else:
                self.fig = plt.figure(igsize=plt.figaspect(self.h_ratio))
                self.ax = self.fig.add_subplot(111, projection='3d')

    def set_decorations(self):

        if self.xlim is not None:
            self.ax.set_xlim(self.xlim)
        if self.ylim is not None:
            self.ax.set_ylim(self.ylim)
        if self.zlim is not None:
            self.ax.set_ylim(self.zlim)

        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)

        self.ax.set_yscale(self.yscale)
        self.ax.set_xscale(self.xscale)
        self.ax.set_zscale(self.zscale)
        self.ax.tick_params(axis='x', labelsize=self.fontsize)
        self.ax.tick_params(axis='y', labelsize=self.fontsize)
        self.ax.tick_params(axis='z', labelsize=self.fontsize)
        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)
        self.ax.set_zlabel(self.zlabel, fontsize=self.fontsize)
        if self.x_formatter:
            self.ax.xaxis.set_major_formatter(self.x_formatter)
        if self.y_formatter:
            self.ax.yaxis.set_major_formatter(self.y_formatter)
        if self.z_formatter:
            self.ax.yaxis.set_major_formatter(self.y_formatter)

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decorations()

        # Categorize by group and marker
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):
                if self.marker != '':
                    marker = self.marker
                else:
                    marker = list(group[self.marker_column])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize

                if self.xmolar:
                    xvals = val_el_to_mol(group[self.xvar])
                else:
                    xvals = group[self.xvar]
                if self.ymolar:
                    yvals = val_el_to_mol(group[self.yvar])
                else:
                    yvals = group[self.yvar]
                if self.zmolar:
                    zvals = val_el_to_mol(group[self.zvar])
                else:
                    zvals = group[self.zvar]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(
                    xvals, yvals, zvals,
                    edgecolors=edge_color,
                    marker=marker, label=label,
                    facecolors=sample_color,
                    s=size,
                    zorder=zorder)

                if self.annotation:
                    for i, sample in xvals.items():
                        self.ax.text(xvals.get(i), yvals.get(i), zvals.get(i),group[self.annotation].get(i), fontsize='xx-small')

        self.plot_arrows()
        self.plot_legend()



from matplotlib.lines import Line2D
from matplotlib.ticker import StrMethodFormatter
from numpy import arange
from georunes.plot.piper.helpers import COS60, SIN60, COS45, project_two_triangles, project_diamond


class CustomFormatter(StrMethodFormatter):
    """
    A child of the StrMethodFormatter class with different formatting if the number is decimal or integer.
    """
    def __init__(self, fmt="%.2f", integer_fmt="%d"):
        super().__init__(fmt)
        self.integer_fmt = integer_fmt

    def format_value(self, value):
        value = round(value, 6)  # A trick to deal with numbers with not exactly precise decimals
        if isinstance(value, float) and value.is_integer():
            return self.integer_fmt % value
        else:
            return self.fmt % value


def line_ions(ax, p1, p2, delta, **kwargs):
    """
    Draws similar lines in the lower fields

    Parameters
    ----------
    ax: PiperAxesSubplot, None
        The subplot to draw on.
    p1: 2-tuple
        The (x,y) starting coordinates
    p2: 2-tuple
        The (x,y) ending coordinates
    delta:
        The space separating the triangles (considering that the width of the diagram is scaled to 1).
    kwargs:
        Any kwargs to pass through to Matplotlib.
    """
    lp1, rp1 = project_two_triangles(p1, delta)
    lp2, rp2 = project_two_triangles(p2, delta)
    ax.add_line(Line2D((lp1[0], lp2[0]), (lp1[1], lp2[1]), **kwargs))
    ax.add_line(Line2D((rp1[0], rp2[0]), (rp1[1], rp2[1]), **kwargs))


def line_diamond(ax, p1, p2, delta, **kwargs):
    """
    Draws similar line in the diamond

    Parameters
    ----------
    ax: PiperAxesSubplot, None
        The subplot to draw on.
    p1: 2-tuple
        The (x,y) starting coordinates
    p2: 2-tuple
        The (x,y) ending coordinates
    delta:
        The space separating the triangles (considering that the width of the diagram is scaled to 1).
    kwargs:
        Any kwargs to pass through to Matplotlib.
    """

    tp1 = project_diamond(p1, delta)
    tp2 = project_diamond(p2, delta)
    ax.add_line(Line2D((tp1[0], tp2[0]), (tp1[1], tp2[1]), **kwargs))


def boundary(ax, delta, axes_colors=None,  **kwargs):
    """
    Plots the boundary of the simplex. Creates and returns matplotlib axis if
    none given.

    Parameters
    ----------
    ax: PiperAxesSubplot, None
        The subplot to draw on.
    delta:
        The space separating the triangles (considering that the width of the diagram is scaled to 1).
    kwargs:
        Any kwargs to pass through to matplotlib.
    axes_colors:
        Option for coloring boundaries different colors.
    """
    pos_top = (0, 1, 0)
    pos_left = (1, 0, 0)
    pos_right = (0, 0, 1)
    dm_left = (0, 0)
    dm_down = (1, 0)
    dm_right = (1, 1)
    dm_up = (0, 1)

    # Set default color as black.
    if axes_colors is None:
        axes_colors = 'black'

    # Horizontal
    line_ions(ax, (*pos_left, *pos_left,), (*pos_right, *pos_right), delta=delta, color=axes_colors, **kwargs)

    # Left
    line_ions(ax, (*pos_top, *pos_top), (*pos_left, *pos_left), delta=delta, color=axes_colors, **kwargs)
    line_diamond(ax, dm_left, dm_up, delta, color=axes_colors, **kwargs)
    line_diamond(ax, dm_down, dm_right, delta, color=axes_colors, **kwargs)

    # Right
    line_ions(ax, (*pos_top, *pos_top), (*pos_right, *pos_right), delta=delta, color=axes_colors, **kwargs)
    line_diamond(ax, dm_left, dm_down, delta=delta, color=axes_colors, **kwargs)
    line_diamond(ax, dm_up, dm_right, delta=delta, color=axes_colors, **kwargs)

    return ax


def gridlines(ax, num_seps=9, delta=0.1, **kwargs):
    """
    Plots grid lines excluding boundary.

    Parameters
    ----------
    ax: PiperAxesSubplot, None
        The subplot to draw on.
    num_seps :
        The number of internal lines by parameter, inside triangles and diamond
    delta:
        The space separating the triangles (considering that the width of the diagram is scaled to 1).
    kwargs:
        Any kwargs to pass through to matplotlib
    """

    if 'linewidth' not in kwargs:
        kwargs["linewidth"] = 0.5
    if 'linestyle' not in kwargs:
        kwargs["linestyle"] = ':'

    interval = 1/(num_seps+1)

    # Draw grid-lines
    # Parallel to horizontal axis
    for i in arange(0, 1, interval):
        # In lower fields
        p1 = (1 - i, i, 0, 1 - i, i, 0)
        p2 = (0, i, 1 - i, 0, i, 1 - i)
        line_ions(ax, p1, p2, delta, **kwargs)

    # Parallel to left and right axes
    for i in arange(0, 1, interval):
        # Left lines
        # In lower fields
        p1 = (1 - i, 0, i, 1 - i, 0, i,)
        p2 = (0, 1 - i, i, 0, 1 - i, i,)
        line_ions(ax, p1, p2, delta, **kwargs)

        # In diamond
        pt1 = (i, 0)
        pt2 = (i, 1)
        line_diamond(ax, pt1, pt2, delta, **kwargs)

        # Right lines
        # In lower fields
        p1 = (i, 0, 1 - i, i, 0, 1 - i,)
        p2 = (i, 1 - i, 0, i, 1 - i, 0,)
        line_ions(ax, p1, p2, delta, **kwargs)
        line_ions(ax, p1, p2, delta, **kwargs)

        # In diamond
        pt1 = (0, i)
        pt2 = (1, i)
        line_diamond(ax, pt1, pt2, delta, **kwargs)

    return ax


def ticks(ax, delta, list_ticks=None, locations=None, num_ticks=6, offset=0.01, clockwise=False, axes_colors=None,
          fontsize=10, tick_formats=None, **kwargs):
    """
    Sets tick marks and labels.

    Parameters
    ----------
    ax: PiperAxesSubplot, None
        The subplot to draw on.
    delta:
        The space separating the triangles (considering that the width of the diagram is scaled to 1).
    list_ticks: list of strings, None
        The tick labels
    locations: list of points, None
        The locations of the ticks
    num_ticks:
        The number of tick on each axis
    offset: float, 0.01
        controls the length of the ticks
    clockwise: bool, False
        Draw ticks marks clockwise or counterclockwise
    axes_colors:
        Option to color ticks differently for each axis,
    fontsize:
        Ticks label fontsize,
    tick_formats: None, Str
        If None, all axes will be labelled with ints.
    kwargs:
        Any kwargs to pass through to matplotlib.

    """
    if tick_formats is None:
        tick_formats = CustomFormatter()
    if axes_colors is None:
        axes_colors = "black"
    length = offset / 3

    if list_ticks and not locations:
        num_ticks = len(list_ticks)
        if num_ticks != 0:
            interval = 1 / (num_ticks - 1)
            locations = arange(0, 1 + interval, interval)

    if not list_ticks:
        interval = 1 / (num_ticks - 1)
        locations = arange(0, 1 + interval, interval)
        list_ticks = locations

    # Right axis
    for index, i in enumerate(locations):
        loc = (0, i, 1 - i)
        if clockwise:
            tick = list_ticks[-(index + 1)]
        else:
            tick = list_ticks[index]

        # Two lower fields
        tl, tr = project_two_triangles((*loc, *loc), delta)
        ax.add_line(
            Line2D((tl[0], tl[0] + length * COS60), (tl[1], tl[1] + length * SIN60), color=axes_colors, **kwargs))
        ax.add_line(
            Line2D((tr[0], tr[0] + length * COS60), (tl[1], tl[1] + length * SIN60), color=axes_colors, **kwargs))

        # Diamond field
        ttl = project_diamond((1 - i, 0), delta)
        ttr = project_diamond((1 - i, 1), delta)
        ax.add_line(Line2D(
            (ttl[0], ttl[0] + length * COS60), (ttl[1], ttl[1] + length * SIN60),
            color=axes_colors, **kwargs))
        ax.add_line(Line2D(
            (ttr[0], ttr[0] + length * COS60), (ttr[1], ttr[1] + length * SIN60),
            color=axes_colors, **kwargs))

        # Text
        pl, pr = project_two_triangles((*loc, *loc), delta)
        pt = project_diamond((1 - i, 1), delta)
        s = tick_formats.format_value(tick * 100)
        ax.text(pl[0] + 0.6 * offset * COS45, pl[1] + 0.6 * offset * COS45, s, horizontalalignment="left",
                color=axes_colors, fontsize=fontsize)
        ax.text(pr[0] + 0.6 * offset * COS45, pr[1] + 0.6 * offset * COS45, s, horizontalalignment="left",
                color=axes_colors, fontsize=fontsize)
        ax.text(pt[0] + 0.6 * offset * COS45, pt[1] + 0.6 * offset * COS45, s, horizontalalignment="left",
                color=axes_colors, fontsize=fontsize)

    # Left axis
    for index, i in enumerate(locations):
        loc = (1 - i, i, 0)
        if clockwise:
            tick = list_ticks[index]
        else:
            tick = list_ticks[-(index + 1)]

        # Two lower fields
        tl, tr = project_two_triangles((*loc, *loc), delta)
        ax.add_line(Line2D(
            (tl[0], tl[0] - length * COS60), (tl[1], tl[1] + length * SIN60),
            color=axes_colors, **kwargs))
        ax.add_line(Line2D(
            (tr[0], tr[0] - length * COS60), (tl[1], tl[1] + length * SIN60),
            color=axes_colors, **kwargs))

        # Diamond field
        ttl = project_diamond((0, 1 - i), delta)
        ttr = project_diamond((1, 1 - i), delta)
        ax.add_line(Line2D(
            (ttl[0], ttl[0] - length * COS60), (ttl[1], ttl[1] + length * SIN60),
            color=axes_colors, **kwargs))
        ax.add_line(Line2D(
            (ttr[0], ttr[0] - length * COS60), (ttr[1], ttr[1] + length * SIN60),
            color=axes_colors, **kwargs))

        # Text
        pl, pr = project_two_triangles((*loc, *loc), delta)
        pt = project_diamond((0, 1 - i), delta)
        s = tick_formats.format_value(tick * 100)
        ax.text(pl[0] - offset * COS45, pl[1], s, horizontalalignment="right",
                color=axes_colors, fontsize=fontsize)
        ax.text(pr[0] - offset * COS45, pr[1], s, horizontalalignment="right",
                color=axes_colors, fontsize=fontsize)
        ax.text(pt[0] - offset * COS45, pt[1], s, horizontalalignment="right",
                color=axes_colors, fontsize=fontsize)

    # Bottom axis
    for index, i in enumerate(locations):
        loc = (1-i, 0, i)
        if clockwise:
            # Right parallel
            tick = list_ticks[-index-1]
        else:
            tick = list_ticks[index]

        tl, tr = project_two_triangles((*loc, *loc), delta)
        ax.add_line(Line2D(
            (tl[0], tl[0]),  (tl[1], tl[1] - length*SIN60),
            color=axes_colors, **kwargs))
        ax.add_line(Line2D(
            (tr[0], tr[0]), (tl[1], tl[1] - length * SIN60),
            color=axes_colors, **kwargs))

        # Text
        pl, pr = project_two_triangles((*loc, *loc), delta)
        s = tick_formats.format_value(tick*100)
        ax.text(pl[0], pl[1]-1.4*offset, s, horizontalalignment="center",
                color=axes_colors, fontsize=fontsize)
        ax.text(pr[0], pr[1]-1.4*offset, s, horizontalalignment="center",
                color=axes_colors, fontsize=fontsize)



import tkinter as tk
from tkinter import ttk


class EditableTreeview(ttk.Treeview):
    def __init__(self, master, data_tab, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.data_tab = data_tab.copy()
        self.bind("<Double-1>", self.on_double_click)
        self.editing_item = None
        self.editing_column = None
        self.entry = None

    def on_double_click(self, event):
        item = self.identify_row(event.y)
        column = self.identify_column(event.x)
        if item and column:
            self.start_editing(item, column)

    def start_editing(self, item, column):
        self.editing_item = item
        self.editing_column = column
        col_index = int(column[1:]) - 1

        # Get the current value
        value = self.item(item, "values")[col_index]

        # Create an Entry widget for editing
        bbox = self.bbox(item, column=column)
        if bbox:
            x, y, width, height = bbox
            self.entry = tk.Entry(self)
            self.entry.place(x=x,
                             y=y,
                             width=width)
            self.entry.insert(0, value)
            self.entry.focus_set()

            # Bind the events for saving and cancelling edit
            self.entry.bind('<Return>', lambda e: self.save_edit(col_index))
            self.entry.bind('<Escape>', self.cancel_edit)

    def finish_item_edition(self):
        if self.editing_item and self.entry:
            new_value = self.entry.get()
            col_index = int(self.editing_column[1:]) - 1
            # column = self.heading(col_index)['text']
            values = list(self.item(self.editing_item, "values"))
            values[col_index] = new_value
            self.item(self.editing_item, values=values)

            # Clean up
            self.entry.destroy()
            self.editing_item = None
            self.editing_column = None
            self.entry = None

    def save_edit(self, col_num):
        if self.editing_item and self.entry:
            new_value = self.entry.get()
            col_index = int(self.editing_column[1:]) - 1
            values = list(self.item(self.editing_item, "values"))
            column = self.heading(col_index)['text']
            self.data_tab.loc[self.editing_item, column] = new_value
            values[col_num] = new_value
            self.item(self.editing_item, values=values)

            # Clean up
            self.entry.destroy()
            self.editing_item = None
            self.editing_column = None
            self.entry = None
            self.event_generate('<<CellUpdated>>', when="tail")

    def cancel_edit(self, event):
        self.entry.destroy()

    def on_focus_out(self, event):
        if self.editing_item:
            self.finish_item_edition()

class EditableFloatTreeview(EditableTreeview):
    def save_edit(self, col_num):
        if self.editing_item and self.entry:
            new_value = self.entry.get()
            col_index = int(self.editing_column[1:]) - 1
            values = list(self.item(self.editing_item, "values"))
            column = self.heading(col_index)['text']

            try:
                float_value = float(new_value)
            except ValueError:
                # If conversion fails, do NOT update and just destroy entry
                self.entry.destroy()
                self.editing_item = None
                self.editing_column = None
                self.entry = None
                return  # exit without update

            self.data_tab.loc[self.editing_item, column] = float_value
            values[col_num] = new_value
            self.item(self.editing_item, values=values)

            # Clean up
            self.entry.destroy()
            self.editing_item = None
            self.editing_column = None
            self.entry = None
            self.event_generate('<<CellUpdated>>', when="tail")




from matplotlib.ticker import LogFormatter


class LogStandardFormatter(LogFormatter):
    def __init__(self, max_num=1000000, min_num=0.0001,
                 **kwargs
                 ):
        self.max_num = max_num
        self.min_num = min_num
        LogFormatter.__init__(self, **kwargs)

    def _num_to_string(self, x, vmin, vmax):
        if x > self.max_num:
            s = '%1.0e' % x
        elif x < self.min_num:
            s = '%1.0e' % x
        else:
            s = self._pprint_val(x, vmax - vmin)
        return s

    def _pprint_val(self, x, d):
        # If the number is not too big and it's an int, format it as an int.
        if abs(x) < self.max_num and x == int(x):
            return '%d' % x
        fmt = ('%1.3e' if x < 1e-3 else
               '%1.3f' if x <= 1 else
               '%1.2f' if x <= 10 else
               '%1.1f' if x <= 1e5 else
               '%1.1e')
        s = fmt % x
        tup = s.split('e')
        if len(tup) == 2:
            mantissa = tup[0].rstrip('0').rstrip('.')
            exponent = int(tup[1])
            if exponent:
                s = '%se%d' % (mantissa, exponent)
            else:
                s = mantissa
        else:
            s = s.rstrip('0').rstrip('.')
        return s



from georunes.petromod.modelers.partition import PartitionModeler


class BatchModeler(PartitionModeler):
    def __init__(self,  bulk_dist_coeffs):
        PartitionModeler.__init__(self, bulk_dist_coeffs)
        self.phases_labels = {"LIQ": "Instantaneous/accumulated melt",
                             "SOL": "Total solid"}
        self.modeler_name = "Batch modeler"

    def ratio_liq(self, liq_fract):
        """
        Calculate the ratio : liquid / initial concentration
        """
        ratio_liq = 1 / (self.bulk_dist_coeffs * (1 - liq_fract) + liq_fract)
        return ratio_liq

    def ratio_sol(self, liq_fract):
        """
        Calculate the ratio : solid / initial concentration
        """
        return self.ratio_liq(liq_fract) * self.bulk_dist_coeffs


class BatchMelting(BatchModeler):
    def __init__(self,  bulk_dist_coeffs):
        BatchModeler.__init__(self, bulk_dist_coeffs)
        self.phases_labels = {"LIQ": "Instantaneous/accumulated melt",
                             "SOL": "Solid residue"}
        self.modeler_name = "Batch melting modeler"
        self.init_name = "Initial solid composition (F=0)"


class BatchCrystallization(BatchModeler):
    def __init__(self,  bulk_dist_coeffs):
        BatchModeler.__init__(self, bulk_dist_coeffs)
        self.phases_labels = {"LIQ": "Instantaneous/accumulated melt",
                             "SOL": "Solid cumulate"}
        self.modeler_name = "Batch crystallization model"
        self.init_name = "Initial liquid composition (F=1)"
        self.f_reverse = True



from itertools import chain
from math import sqrt
import numpy as np
from georunes.tools.chemistry import el_molar_mass as _mm

COS60 = 1 / 2
SIN60 = sqrt(3) / 2
COS45 = SIN45 = sqrt(2) / 2

ratio_meq = {
    "Ca2+": 2 / _mm["Ca"],
    "Mg2+": 2 / _mm["Mg"],
    "Na+": 1 / _mm["Na"],
    "K+": 1 / _mm["K"],
    "HCO3-": 1 / (_mm["H"] + _mm["C"] + _mm["O"] * 3),
    "CO3-": 1 / (_mm["C"] + _mm["O"] * 3),
    "SO42-": 2 / (_mm["S"] + _mm["O"] * 4),
    "Cl-": 1 / _mm["Cl"],
}


def ions_mgl_to_meq(p):
    _ca, _mg, _na, _k, _hco, _co, _so, _cl = p
    ca = cations_mgl_to_meq((_ca, _mg, _na, _k))
    an = anions_mgl_to_meq((_hco, _co, _so, _cl))
    return *ca, *an


def cations_mgl_to_meq(cations):
    _ca, _mg, _na, _k = cations
    ca = _ca * ratio_meq["Ca2+"]
    mg = _mg * ratio_meq["Mg2+"]
    nak = _na * ratio_meq["Na+"] + _k * ratio_meq["K+"]
    return ca, mg, nak


def anions_mgl_to_meq(anions):
    _hco, _co, _so, _cl = anions
    hco = _hco * ratio_meq["HCO3-"] + _co * ratio_meq["CO3-"]
    so = _so * ratio_meq["SO42-"]
    cl = _cl * ratio_meq["Cl-"]
    return hco, so, cl


def camg_mgl_to_meq(_ca, _mg):
    ca = _ca * ratio_meq["Ca2+"]
    mg = _mg * ratio_meq["Mg2+"]
    return ca + mg


def socl_mgl_to_meq(_so, _cl):
    so = _so * ratio_meq["SO42-"]
    cl = _cl * ratio_meq["Cl-"]
    return so + cl


def diamond_ox(delta):
    """ Get the X position of the left vertex of the diamond """
    width = (1 - delta) / 2
    return width / 2 + delta * COS60


def diamond_oy(delta):
    """ Get the Y position of the left vertex of the diamond """
    width = (1 - delta) / 2
    return width * SIN60 + delta * SIN60


def unzip(list_tuples):
    """[(a1, b1), ..., (an, bn)] ----> ([a1, ..., an], [b1, ..., bn])"""
    return list(zip(*list_tuples))


def project_sequence(seq, delta):
    """
    Projects a point or sequence of points using `project_point`.
    Returns the points in coordinates as two lists xs and ys
    """
    # Projection -> [ [[x1_cation, y1_cation], [x1_anion, y1_anion], [x1_diamond, y1_diamond]],
    #                 [[x2_cation, y2_cation], [x2_anion, y2_anion], [x2_diamond, y2_diamond]], ... ]
    projected = [project_points_piper(p, delta) for p in seq]

    # Group by coordinates
    #           -> [ [(x1_cation, x1_anion, x1_diamond), (y1_cation, y1_anion, y1_diamond)],
    #                [(x2_cation, x2_anion, x2_diamond), (y2_cation, y2_anion, y2_diamond)], ... ]
    list_tripoints = [unzip(pts) for pts in projected]

    # Merge in unique list for each coordinate
    #            -> [[x1_cation, x1_anion, x1_diamond, x2_cation, x2_anion, x2_diamond, ...],
    #                [y1_cation, y1_anion, y1_diamond, y2_cation, , y2_anion, y2_diamond, ...]]
    merged = [list(chain(*items)) for items in zip(*list_tripoints)]
    return merged


def project_points_piper(p, delta):
    """ Maps p = (ca, mg, na, k, hco3, co3, so, cl) coordinates to planar simplex. """
    width = (1 - delta) / 2
    ca, mg, nak, hco, so, cl = ions_mgl_to_meq(p)

    val_left = nak / (mg + ca + nak)
    val_right = (so + cl) / (so + hco + cl)

    tx = diamond_ox(delta) + val_left * width * COS60 + val_right * width * COS60
    ty = diamond_oy(delta) + width * val_left * (-1) * SIN60 + width * val_right * SIN60
    lx = width * (COS60 * (2 * nak + mg) / (mg + ca + nak))
    ly = width * SIN60 * mg / (mg + ca + nak)
    rx = width + delta + width * (COS60 * (2 * cl + so) / (so + hco + cl))
    ry = width * SIN60 * so / (so + hco + cl)

    return [[lx, ly], [rx, ry], [tx, ty]]


def project_ions_diamond(p, delta, drift=None):
    """ Get (ca,mg,so,cl) and return (x,y) position """
    if drift is None:
        drift = (0, 0)
    _ca, _mg, _so, _cl = p
    camg = camg_mgl_to_meq(_ca, _mg)
    socl = socl_mgl_to_meq(_so, _cl)
    width = (1 - delta) / 2
    tx = diamond_ox(delta) + camg * width * COS60 + socl * width * COS60
    ty = diamond_oy(delta) + camg * width * (-1) * SIN60 + socl * width * SIN60
    return [tx + drift[0], ty + drift[1]]


def project_two_triangles(p, delta, drift=None):
    """
    Maps p = (llf, ltp, lrt, rlf, rtp, rrt) to planar simplex in the anions and cations triangles,
    returns 2 couples (x,y)
    """
    if drift is None:
        drift = (0, 0)
    llf, ltp, lrt, rlf, rtp, rrt = p
    return [project_left_triangle(( llf, ltp, lrt), delta=delta, drift=drift),
            project_right_triangle((rlf, rtp, rrt), delta=delta, drift=drift)]


def project_left_triangle(p, delta, drift=None):
    """ Maps p = (left, top, right) composition to planar simplex in the cations triangle, returns a couple (x,y) """
    if drift is None:
        drift = (0, 0)
    lf, tp, rt = p
    width = (1 - delta) / 2
    lx = width * (0.5 * (2 * rt + tp) / (tp + lf + rt))
    ly = width * (sqrt(3) / 2) * tp / (tp + lf + rt)
    return [lx + drift[0], ly + drift[1]]


def project_right_triangle(p, delta, drift=None):
    """ Maps p = (left, top, right)  to planar simplex in the anions triangle, returns a couple (x,y) """
    if drift is None:
        drift = (0, 0)
    lf, tp, rt = p
    width = (1 - delta) / 2
    rx = width + delta + width * (0.5 * (2 * rt + tp) / (tp + lf + rt))
    ry = width * (sqrt(3) / 2) * tp / (tp + lf + rt)
    return [rx + drift[0], ry + drift[1]]


def project_diamond(p, delta, drift=None):
    """ Get (ca,mg,so,cl) and return (x,y) position """
    if drift is None:
        drift = (0, 0)
    down_left, down_right = p
    width = (1 - delta) / 2
    tx = diamond_ox(delta) + down_left * width * COS60 + down_right * width * COS60
    ty = diamond_oy(delta) + down_left * width * (-1) * SIN60 + down_right * width * SIN60
    return [tx + drift[0], ty + drift[1]]


def split_by_field(xs, ys):
    """ Get x and y list of points, return 3 lists for any a field (left, right and diamond). """
    x3 = np.array_split(xs, 3)
    xsp = list(zip(*x3))
    y3 = np.array_split(ys, 3)
    ysp = list(zip(*y3))
    return list(zip(xsp, ysp))



from georunes.petromod.modelers.batch import BatchModeler
from georunes.petromod.modelers.partition import get_non_modal_dist_coeffs


class BatchNonModal(BatchModeler):  # todo implement
    def __init__(self, bulk_dist_coeffs, melt_propotions, part_coeffs):
        BatchModeler.__init__(self, bulk_dist_coeffs)
        self.nmod_dist_coeffs = get_non_modal_dist_coeffs(melt_propotions, part_coeffs)

    def ratio_liq(self, liq_fract):
        """
        Calculate the ratio : liquid / initial concentration
        """
        ratio_liq = 1 / (self.bulk_dist_coeffs + liq_fract * (1 - self.nmod_dist_coeffs))
        return ratio_liq

    def ratio_sol(self, liq_fract):
        """
        Calculate the ratio : solid / initial concentration
        """
        ratio_sol = self.bulk_dist_coeffs.copy()
        for index, row in ratio_sol.items():
            _part = (self.bulk_dist_coeffs[index] - liq_fract * self.nmod_dist_coeffs[index]) / (1 - liq_fract)
            ratio_sol[index] = _part / (self.bulk_dist_coeffs[index] + liq_fract * (1 - self.nmod_dist_coeffs[index]))
        return ratio_sol



import numpy as np
from matplotlib.colors import to_rgba

from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer, ArrowDrawer
from georunes.tools.chemistry import val_ox_to_mc
from georunes.tools.language import get_translator
from georunes.tools.plotting import get_spline


# Debon, F. and Le Fort, P., 1983. A chemical–mineralogical classification of common plutonic rocks and associations.
# Earth and Environmental Science Transactions of the Royal Society of Edinburgh, 73(3), pp.135-149.

class DiagramBA(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title="B-A multicationic classification (Debon and Le Fort 1983)",
                 annotation=None, decor_set="Debon",
                 legend_loc="upper right",
                 alpha_color=0.4, alpha_edge_color=0.8,
                 **kwargs):
        DiagramBase.__init__(self, datasource=datasource, title=title, legend_loc=legend_loc, **kwargs)

        self.decor_set = decor_set
        self.annotation = annotation
        self.xlabel = "B = Fe + Mg + Ti"
        self.ylabel = "A = Al - (Na+ K + 2Ca)"
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color

    def set_decoration(self):
        _ = get_translator(self.lang_cfg)
        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        self.ax.set_xlim(0, 200)
        self.ax.set_ylim(-100, 200)

        xax = (0, 200)
        yax = (0, 0)

        self.ax.plot(xax, yax, color=self.decor_line_col, linewidth=1.0)

        if self.decor_set is "Villaseca":

            lx1 = np.array([0, 13, 29, 37, 36, 32, 30])
            ly1 = np.array([100, 102, 98, 87, 54, 25, 0])

            lx2 = np.array([29, 45, 57, 78])
            ly2 = np.array([98, 148, 176, 200])

            lx3 = np.array([36, 92, 145, 167, 200])
            ly3 = np.array([54, 44, 39, 42, 53])

            lx4 = np.array([32, 57, 82, 105, 130])
            ly4 = np.array([25, 21, 14, 8, 0])

            x1, y1 = get_spline(lx1, ly1)
            x2, y2 = get_spline(lx2, ly2)
            x3, y3 = get_spline(lx3, ly3)
            x4, y4 = get_spline(lx4, ly4)

            self.ax.plot(x1, y1, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(x2, y2, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(x3, y3, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(x4, y4, color=self.decor_line_col, linewidth=0.6)

            self.ax.text(75, 110, "hP", size='x-small')
            self.ax.text(55, 40, "mP", size='x-small')
            self.ax.text(50, 10, "lP", size='x-small')
            self.ax.text(18, 60, _("felsic"), size='x-small')

        else:  # Default decor_set is from Debon and Le Fort 1983

            lx1 = np.array([0, 420])
            ly1 = np.array([0, -885])

            lx2 = np.array([0, 470])
            ly2 = np.array([0, -225])

            lx3 = np.array([0, 1500])
            ly3 = np.array([0, 0])

            lx4 = np.array([0, 600])
            ly4 = np.array([0, 250])

            lx5 = np.array([0, 435])
            ly5 = np.array([0, 500])

            self.ax.plot(lx1, ly1, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx2, ly2, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx3, ly3, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx4, ly4, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx5, ly5, color=self.decor_line_col, linewidth=0.6)

            self.ax.text(100, 150, "I", size='x-small')
            self.ax.text(175, 150, "II", size='x-small')
            self.ax.text(180, 25, "III", size='x-small')
            self.ax.text(180, -25, "IV", size='x-small')
            self.ax.text(75, -65, "V", size='x-small')
            self.ax.text(20, -75, "VI", size='x-small')

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:
            if self.is_group_allowed(name):

                al_mc = val_ox_to_mc(group["Al2O3"])
                na_mc = val_ox_to_mc(group["Na2O"])
                k_mc = val_ox_to_mc(group["K2O"])
                ca_mc = val_ox_to_mc(group["CaO"])
                fe_mc = val_ox_to_mc(group["Fe2O3"])
                mg_mc = val_ox_to_mc(group["MgO"])
                ti_mc = val_ox_to_mc(group["TiO2"])

                param_A = al_mc - na_mc - k_mc - 2 * ca_mc
                param_B = fe_mc + mg_mc + ti_mc

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(param_B, param_A, edgecolors=edge_color,
                                marker=list(group[self.marker_column])[0], label=label, facecolors=sample_color,
                                s=self.markersize,
                                zorder=zorder)

                if self.annotation:
                    for i, sample in param_B.items():
                        self.ax.annotate(group[self.annotation].get(i), (param_B.get(i), param_A[i]),
                                         fontsize='xx-small')

        self.plot_arrows()
        self.plot_legend()



import pandas as pd

from georunes.tools.chemistry import molar_mass, el_molar_mass, number_cations_per_oxygen, get_cation, \
    number_oxygen_in_oxide
from georunes.modmin.optim.base import BaseOptimizer

supported_oxides = ['SiO2', 'TiO2', 'CO2', 'Al2O3', 'Cr2O3', 'Fe2O3', 'FeO', 'Mn2O3', 'MnO', 'MgO', 'NiO', 'CaO', 'BaO',
                    'SrO', 'Li2O', 'Na2O', 'K2O', 'Rb2O', 'P2O5', 'H2O']
supported_anions = ['F', 'Cl']

custom_cation = {
    "Fe2O3": "Fe3+",
    "FeO": "Fe2+",
    "Mn2O3": "Mn3+",
    "MnO": "Mn2+"
}


class APFUCalc(BaseOptimizer):
    def __init__(self, **kwargs):
        BaseOptimizer.__init__(self, **kwargs)
        self.notif = ">>>>>> Atoms per formula unit (APFU) calculation"

    def compute(self, raw_data, skip_cols, oxygen_number=12, oxygen_equiv=False, to_round=3,
                ignored_columns=('LOI', 'H2O+', 'Total'), force_cation_number=None):
        if ignored_columns is None:
            ignored_columns = []

        # 1 / Preparing data
        if self.verbose > 1: print("Step 1 - Prepare data")

        data = raw_data.iloc[:, :skip_cols].copy()
        raw_data_keys = raw_data.iloc[:, skip_cols:].keys()
        supported_oxel = [*supported_oxides, *supported_anions]  # Limit the list of usable oxide
        for key in raw_data_keys:
            if key in supported_oxel and key not in ignored_columns:
                data[key] = raw_data[key]
        data = data.fillna(0)
        data_keys = data.keys()
        list_oxel = [ox for ox in data_keys if ox in supported_oxel]
        list_ox = [ox for ox in list_oxel if ox in supported_oxides]

        filtered_columns = [oxel for oxel in raw_data_keys if oxel not in data_keys]
        if self.verbose: print('Ignored columns :', filtered_columns)

        moles_oxides = data.iloc[:, :skip_cols].copy()
        moles_oxygen = data.iloc[:, :skip_cols].copy()
        ox_normalized = data.iloc[:, :skip_cols].copy()
        moles_cations = data.iloc[:, :skip_cols].copy()
        suppl = data.iloc[:, :skip_cols].copy()
        suppl["sum_cations"] = 0
        suppl["sum_anions"] = 0

        # 2 / Weight percent to moles number in oxides
        if self.verbose > 1: print("Step 2 - Calculate number of moles from weight percent of each oxide")
        for oxel in list_oxel:
            if oxel in supported_anions:
                moles_oxides[oxel] = data[oxel] / el_molar_mass[oxel]
            else:
                moles_oxides[oxel] = data[oxel] / molar_mass[oxel]

        # 3 / Moles number to oxygen moles number in oxides
        if self.verbose > 1: print("Step 3 - Calculate number of moles of oxygen in each oxide")

        for oxel in list_oxel:
            if oxel == 'H2O':
                moles_oxygen['H2O'] = moles_oxides['H2O']
            elif oxel in list_ox:
                moles_oxygen[oxel] = moles_oxides[oxel] * number_oxygen_in_oxide(oxel)
            else:
                moles_oxygen[oxel] = moles_oxides[oxel]

        # 4 / Normalization to the number of oxygen per formula
        if self.verbose > 1: print("Step 4 - Normalization to the number of oxygen per formula")

        _o = 'O'
        sum_o = moles_oxygen.iloc[:, skip_cols:].sum(axis=1)
        if oxygen_equiv:
            for element in supported_anions:
                if element in list_oxel:
                    sum_o = sum_o - moles_oxygen[element]/2  # 2 F,Cl for one O
                    _o = _o+'+'+str(element)

        oxygen_factor = oxygen_number / sum_o
        suppl["O_factor"] = oxygen_factor
        for oxel in list_oxel:
            ox_normalized[oxel] = moles_oxygen[oxel] * oxygen_factor

        # 5 / Normalized oxygen numbers to cations number
        if self.verbose > 1: print("Step 5 - Calculation of each cation number according to the normalized oxygen")

        for oxel in list_oxel:
            if oxel == 'H2O':
                moles_cations['OH'] = ox_normalized[oxel] * 2
                suppl["sum_anions"] = suppl["sum_anions"] + moles_cations['OH']
            elif oxel in supported_anions:
                moles_cations[oxel] = ox_normalized[oxel]
                suppl["sum_anions"] = suppl["sum_anions"] + moles_cations[oxel]
            elif oxel in custom_cation.keys():
                cat = custom_cation[oxel]
                moles_cations[cat] = ox_normalized[oxel] * number_cations_per_oxygen(oxel)
                suppl["sum_cations"] = suppl["sum_cations"] + moles_cations[cat]
            else:
                cat = get_cation(oxel)
                moles_cations[cat] = ox_normalized[oxel] * number_cations_per_oxygen(oxel)
                if cat not in ('P', 'C'):
                    suppl["sum_cations"] = suppl["sum_cations"] + moles_cations[cat]
        moles_cations[_o] = oxygen_number

        # If the cations number is fixed
        if isinstance(force_cation_number, int) and force_cation_number > 0:
            if self.verbose > 1: print("Number of cations forced to " + str(force_cation_number))
            moles_cations_cat_fixed = data.iloc[:, :skip_cols].copy()
            sum_cat = pd.Series([0] * len(moles_cations_cat_fixed), dtype=float, )
            list_atoms = moles_cations.iloc[:, skip_cols:].keys()
            for cat in list_atoms:
                if cat not in ('P', 'C', 'OH', 'O', 'O+F', 'O+Cl', 'O+F+Cl', 'O+Cl+F', *supported_anions):
                    sum_cat = sum_cat + moles_cations[cat]
            cations_factor = force_cation_number / sum_cat

            for cat in list_atoms:
                moles_cations_cat_fixed[cat] = moles_cations[cat] * cations_factor
            suppl["sum_anions_CF"] = suppl["sum_anions"] * cations_factor
            suppl["sum_cations_CF"] = suppl["sum_cations"] * cations_factor
            suppl['cations_factor'] = cations_factor
            moles_cations_cat_fixed = moles_cations_cat_fixed.round(to_round)

        # End
        moles_cations = moles_cations.round(to_round)

        suppl = suppl.round(to_round)
        suppl = suppl.fillna(0)
        if self.verbose:
            print(">>> Atoms")
            print(moles_cations.to_string())
        if self.verbose > 1:
            if isinstance(force_cation_number, int) and force_cation_number > 0:
                print(">>> Atoms (with cations normalized to " + str(force_cation_number) + ")")
                print(moles_cations_cat_fixed.to_string())
            print(">>> Supplementary data")
            print(suppl.to_string())

        return moles_cations, suppl

    def compute_oxygen_basis(self, raw_data, skip_cols, oxygen_number, **kwargs):
        return self.compute(raw_data, skip_cols, oxygen_number=oxygen_number, **kwargs)

    def compute_cation_basis(self, raw_data, skip_cols,  force_cation_number, oxygen_number=12, **kwargs):
        return self.compute(raw_data, skip_cols, force_cation_number=force_cation_number, oxygen_number=oxygen_number,
                            **kwargs)



from matplotlib.colors import to_rgba

from georunes.plot.helpers import ArrowDrawerTernary
from georunes.plot.ternary.ternbase import DiagramTernaryBase
from georunes.tools.language import get_translator
from georunes.tools.plotting import normalize_marker_size


# Barker, F., 1979. Trondhjemite: definition, environment and hypotheses of origin. In Developments in petrology (
# Vol. 6, pp. 1-12). Elsevier.
# O'Connor, J.T., 1965. A classification of quartz-rich igneous rocks based on feldspar ratios.
# US Geol. Surv. Prof. Paper, 82.

class DiagramAnAbOr(DiagramTernaryBase, ArrowDrawerTernary):
    def __init__(self, datasource, title="Ternary feldspar plot (anorthite, albite and orthoclase)",
                 decor_set="Barker79", padding={"bottom": 0.124}, **kwargs):

        if decor_set not in ("blank", "OConnor65", "Barker79"):
            raise ValueError("Parameter 'decor_set' must be 'blank', 'OConnor65' or 'Barker79'")

        DiagramTernaryBase.__init__(self, datasource=datasource, title=title,
                                    top_var='An', left_var='Ab', right_var='Or',
                                    top_label='An', left_label='Ab', right_label='Or',
                                    tscale='linear', left_scale='linear', right_scale='linear',
                                    padding=padding,
                                    **kwargs)

        self.decor_set = decor_set

    def set_decoration(self, ):
        _ = get_translator(self.lang_cfg)

        # Order for points : Or, An, Ab
        if self.decor_set == "blank":
            pass

        elif self.decor_set == "Barker79":

            p1 = (0, 30, 70)
            p2 = (24, 18, 58)

            p21 = (24, 18, 58)
            p22 = (34, 16, 50)

            p31 = (20, 20, 60)
            p32 = (20, 50, 30)

            p41 = (34, 16, 50)
            p42 = (34, 42, 24)

            p51 = (24, 18, 58)
            p52 = (30, 00, 70)

            self.tax.line(p1, p2, linewidth=1, color=self.decor_line_col)
            self.tax.line(p21, p22, linewidth=1, color=self.decor_line_col)
            self.tax.line(p31, p32, linewidth=1, color=self.decor_line_col)
            self.tax.line(p41, p42, linewidth=1, color=self.decor_line_col)
            self.tax.line(p51, p52, linewidth=1, color=self.decor_line_col)

            self.tax.annotate(_("Granite"), (30, 10,), fontsize="small")
            self.tax.annotate(_("Tonalite"), (55 * 0.08, 45, 55 * 0.92), fontsize="small", rotation=60)
            self.tax.annotate(_("Granodiorite"), (55 * 0.5, 45, 55 * 0.5), fontsize="small", rotation=60)
            self.tax.annotate(_("Trondhjemite"), (5, 5,), fontsize="small")

        elif self.decor_set == "OConnor65":
            p1 = (20, 0, 80)
            p2 = (20, 20, 60)

            p21 = (0, 25, 75)
            p22 = (60, 10, 30)

            p31 = (60, 10, 30)
            p32 = (100, 0, 0)
            p41 = (60, 10, 30)
            p42 = (60, 20, 20)
            p51 = (35, 16.25, 48.5)
            p52 = (35, 32.5, 32.5)
            p61 = (10, 22., 5, 67.5)
            p62 = (10, 45, 45)

            self.tax.line(p1, p2, linewidth=1, color=self.decor_line_col)
            self.tax.line(p21, p22, linewidth=1, color=self.decor_line_col)
            self.tax.line(p31, p32, linewidth=1, color=self.decor_line_col)
            self.tax.line(p41, p42, linewidth=1, color=self.decor_line_col)
            self.tax.line(p51, p52, linewidth=1, color=self.decor_line_col)
            self.tax.line(p61, p62, linewidth=1, color=self.decor_line_col)
            self.tax.annotate(_("Trondhjemite"), (5, 5,), fontsize="small")
            self.tax.annotate(_("Granite"), (35, 5,), fontsize="small")
            self.tax.annotate(_("Tonalite"), (4, 35,), fontsize="small", rotation=60)
            self.tax.annotate(_("Granodiorite"), (20, 35,), fontsize="small", rotation=60)
            self.tax.annotate(_("Quartz monzonite"), (45, 25,), fontsize="small", rotation=60)

    def plot(self):
        DiagramTernaryBase.plot_config(self, no_ticks=True)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):
                var_sum = group[self.top_var] + group[self.left_var] + group[self.right_var]
                norm_z = 100 * group[self.left_var] / var_sum
                norm_x = 100 * group[self.right_var] / var_sum
                norm_y = 100 * group[self.top_var] / var_sum

                if self.marker != '':
                    marker = self.marker
                else:
                    marker = list(group[self.marker_column])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize
                # Convert points to tuples

                points = [(norm_x.get(i), norm_y.get(i), norm_z.get(i)) for i, sample in
                          group[self.top_var].items()]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.tax.scatter(points, edgecolors=edge_color,
                                marker=marker, label=label, facecolors=sample_color, s=size,
                                zorder=zorder)

                if self.annotation:
                    for i, sample in group[self.top_var].items():
                        self.tax.annotate(group[self.annotation].get(i), (norm_x.get(i), norm_y.get(i)),
                                          fontsize='xx-small')

        self.plot_legend()



import re

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources
import gettext


def format_chemical_formula(formula):
    items = re.findall(r"[A-Za-z() -*%.+*/]+|[0-9]+", formula, re.I)
    if items:
        for idx in range(len(items)):
            if items[idx].isnumeric():
                items[idx] = "$_{" + items[idx] + "}$"
        return ''.join(items)
    return formula


def get_translator(lang_cfg):
    if lang_cfg is not None and lang_cfg['lang'] is not None:
        lang = lang_cfg['lang']
        if 'domain' in lang_cfg.keys():
            domain = lang_cfg['domain']
        else:
            domain = 'gg'
        if 'locales' in lang_cfg.keys():
            locales = lang_cfg['locales']
        else:
            locales = 'locales'

        locales_path = pkg_resources.files(locales).joinpath("")
        el = gettext.translation(domain, localedir=str(locales_path), languages=[lang, ])
        el.install()
        return el.gettext

    return gettext.gettext



from abc import abstractmethod, ABC
import pandas as pd


def initial_concentration_as_series(list_elements, list_concentrations):
    return pd.Series(list_concentrations, index=list_elements)


def compute_bulk_coeffs(modal_mineralogy, part_coeffs):
    bulk_dist_coeffs = pd.DataFrame(index=part_coeffs.index)
    df_prop = pd.Series(modal_mineralogy)
    for index, row in part_coeffs.iterrows():
        res = row * df_prop
        val = res.sum() / 100
        bulk_dist_coeffs.at[index, 'D'] = val
    return bulk_dist_coeffs['D']


def get_non_modal_dist_coeffs(melt_proportions, part_coeffs):
    if sum(melt_proportions.values()) != 1:
        raise Exception("The sum of the melt proportions must be 1.")
    part_coeffs_minerals = part_coeffs.keys()

    # Verify the minerals provided with melt proportions
    missing_minerals = []
    for mineral in melt_proportions:
        if mineral not in part_coeffs_minerals:
            missing_minerals.append(mineral)
    if missing_minerals:
        notification = "Minerals missing in the dataset: " + str(missing_minerals)
        raise Exception(notification)

    p = pd.DataFrame(index=part_coeffs.index)
    for element, row in part_coeffs.iterrows():
        tot = 0
        for mineral, prop in melt_proportions.items():
            tot += row[mineral] * prop
        p.at[element, 'P'] = tot

    return p['P']


class PartitionModeler(ABC):
    def __init__(self, bulk_dist_coeffs):
        self.phases_labels = None
        self.bulk_dist_coeffs = bulk_dist_coeffs
        self.modeler_name = "Default partition modeler"
        self.f_reverse = False

    def set_bulk_dist_coeffs(self, bulk_dist_coeffs):
        self.bulk_dist_coeffs = bulk_dist_coeffs

    def list_elements(self):
        keys = self.bulk_dist_coeffs.index.tolist()
        return keys

    @abstractmethod
    def ratio_liq(self, liq_fract):
        """
        Calculate the ratio : liquid / initial concentration
        """
        pass

    @abstractmethod
    def ratio_sol(self, liq_fract):
        """
        Calculate the ratio : solid / initial concentration
        """
        return self.ratio_liq(liq_fract) * self.bulk_dist_coeffs

    def concentration_liq_el(self, el, liq_fract, initial_c0):
        """
        Calculate the ratio for an element : liquid / initial concentration
        """
        c_liq_el = initial_c0[el] * self.ratio_liq(liq_fract)[el]
        return c_liq_el

    def concentration_sol_el(self, el, liq_fract, initial_c0):
        """
        Calculate the ratio for an element: solid / initial concentration
        """
        c_sol_el = initial_c0[el] * self.ratio_sol(liq_fract)[el]
        return c_sol_el

    def concentration_liq_els(self, list_elements, liq_fract, initial_c0):
        """
        Calculate the ratio for a list of elements : liquid / initial concentration
        """
        return [self.concentration_liq_el(el, liq_fract, initial_c0) for el in list_elements]

    def concentration_sol_els(self, list_elements, liq_fract, initial_c0):
        """
        Calculate the ratio for a list of elements: solid / initial concentration
        """
        return [self.concentration_sol_el(el, liq_fract, initial_c0) for el in list_elements]

    def get_phase_concentration_func(self, phase):
        if phase == 'LIQ':
            return self.concentration_liq_el
        elif phase == 'SOL':
            return self.concentration_sol_el
        return None

    def get_phases_labels(self):
        return self.phases_labels



import tkinter as tk


class EntryWithCallback(tk.Entry):
    def __init__(self, parent=None, callback=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.callback = callback
        self.var = tk.StringVar()
        self.var.trace_add("write", self._on_change)  # Attach the callback
        self.config(textvariable=self.var)
        self.callback_enabled = False

    def _on_change(self, *args):
        if self.callback_enabled and self.callback:
            self.callback()

    def disable_callback(self):
        self.callback_enabled = False

    def enable_callback(self):
        self.callback_enabled = True



import numpy as np
from matplotlib.colors import to_rgba

from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer, ArrowDrawer
from georunes.tools.language import get_translator


# Pearce, J.A., Harris, N.B.W. and Tindle, A.G., 1984. Trace Element Discrimination Diagrams for the Tectonic
# Interpretation of Granitic Rocks. Journal of Petrology, 25(4), pp.956-983.

class PearceBase(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource,
                 x_formatter=None, y_formatter=None,
                 alpha_color=0.4, alpha_edge_color=0.8,
                 padding=None,
                 **kwargs
                 ):
        config_padding = {"left": 0.16}
        if padding:
            config_padding.update(padding)

        DiagramBase.__init__(self, datasource=datasource, padding=config_padding, **kwargs)
        self.ax.set_xscale('log')
        self.ax.set_yscale('log')
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color
        self.x_formatter = x_formatter
        self.y_formatter = y_formatter

    def set_decorations(self):
        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        if self.xlim is not None:
            self.ax.set_xlim(self.xlim)
        if self.ylim is not None:
            self.ax.set_ylim(self.ylim)

        if self.x_formatter:
            self.ax.xaxis.set_major_formatter(self.x_formatter)
        if self.y_formatter:
            self.ax.yaxis.set_major_formatter(self.y_formatter)


class DiagramPearceRYN(PearceBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title="Pearce : Rb vs Y + Nb",
                 xlim=(2, 2000), ylim=(1, 2000),
                 **kwargs
                 ):
        PearceBase.__init__(self, datasource=datasource, title=title, **kwargs)
        self.xlim = xlim
        self.ylim = ylim
        self.xlabel = "Y + Nb (ppm)"
        self.ylabel = "Rb (ppm)"

    def set_decorations(self):
        PearceBase.set_decorations(self)
        _ = get_translator(self.lang_cfg)

        lx1 = np.array([2, 55])
        ly1 = np.array([80, 300])

        lx2 = np.array([55, 400])
        ly2 = np.array([300, 2000])

        lx3 = np.array([55, 51.5])
        ly3 = np.array([300, 8])

        lx4 = np.array([51.5, 50])
        ly4 = np.array([8, 1])

        lx5 = np.array([51.5, 2000])
        ly5 = np.array([8, 400])

        self.ax.plot(lx1, ly1, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx2, ly2, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx3, ly3, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx4, ly4, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx5, ly5, self.decor_line_col, linewidth=0.6)

        self.ax.text(6.3, 756, _("Syn-COLG"), color=self.decor_text_col)
        self.ax.text(210, 260, _("WPG"), color=self.decor_text_col)
        self.ax.text(4.36, 2.6, _("VAG"), color=self.decor_text_col)
        self.ax.text(277, 5, _("ORG"), color=self.decor_text_col)

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decorations()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.exclude_groups and name not in self.exclude_groups:

                vx = group["Nb"] + group["Y"]
                vy = group["Rb"]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(vx, vy, edgecolors=edge_color,
                                marker=list(group[self.marker_column])[0], label=label,
                                facecolors=sample_color,
                                s=self.markersize,
                                zorder=zorder)
        self.plot_arrows()
        self.plot_legend()
        self.adjust_padding()


class DiagramPearceRYT(PearceBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title="Pearce : Rb vs Yb + Ta",
                 xlim=(0.5, 200), ylim=(1, 2000),
                 **kwargs
                 ):
        PearceBase.__init__(self, datasource=datasource, title=title, **kwargs)
        self.xlim = xlim
        self.ylim = ylim
        self.xlabel = "Yb + Ta (ppm)"
        self.ylabel = "Rb (ppm)"

    def set_decorations(self):
        PearceBase.set_decorations(self)
        _ = get_translator(self.lang_cfg)

        lx1 = np.array([6, 50])
        ly1 = np.array([200, 2000])

        lx2 = np.array([6, 6])
        ly2 = np.array([200, 8])

        lx3 = np.array([6, 6])
        ly3 = np.array([8, 1])

        lx4 = np.array([6, 200])
        ly4 = np.array([8, 400])

        lx5 = np.array([0.5, 6])
        ly5 = np.array([140, 200])

        self.ax.plot(lx1, ly1, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx2, ly2, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx3, ly3, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx4, ly4, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx5, ly5, self.decor_line_col, linewidth=0.6)

        self.ax.text(5.5, 1210, _("Syn-COLG"), color=self.decor_text_col)
        self.ax.text(54, 479, _("WPG"), color=self.decor_text_col)
        self.ax.text(2, 1.4, _("VAG"), color=self.decor_text_col)
        self.ax.text(24, 1.4, _("ORG"), color=self.decor_text_col)

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decorations()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.exclude_groups and name not in self.exclude_groups:

                vx = group["Ta"] + group["Yb"]
                vy = group["Rb"]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(vx, vy, edgecolors=edge_color,
                                marker=list(group[self.marker_column])[0], label=label, facecolors=sample_color,
                                s=self.markersize,
                                zorder=zorder)

        self.plot_arrows()
        self.plot_legend()


class DiagramPearceNY(PearceBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title="Pearce : Nb vs Y",
                 xlim=(1, 2000), ylim=(1, 2000),
                 **kwargs
                 ):
        PearceBase.__init__(self, datasource=datasource, title=title, **kwargs)
        self.xlim = xlim
        self.ylim = ylim
        self.xlabel = "Y (ppm)"
        self.ylabel = "Nb (ppm)"

    def set_decorations(self):
        PearceBase.set_decorations(self)
        _ = get_translator(self.lang_cfg)

        lx1 = np.array([40, 50])
        ly1 = np.array([1, 10])

        lx2 = np.array([50, 1000])
        ly2 = np.array([10, 100])

        lx3 = np.array([1, 50])
        ly3 = np.array([2000, 10])

        lx4 = np.array([25, 1000])
        ly4 = np.array([25, 400])

        self.ax.plot(lx1, ly1, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx2, ly2, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx3, ly3, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx4, ly4, color=self.decor_line_col, linewidth=0.6, linestyle='--')

        self.ax.text(2.3, 16.9, _("Syn-COLG") + " +\n" + _("VAG"), color=self.decor_text_col)
        self.ax.text(57, 301, _("WPG"), color=self.decor_text_col)
        self.ax.text(277, 5, _("ORG"), color=self.decor_text_col)

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decorations()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.exclude_groups and name not in self.exclude_groups:

                vx = group["Y"]
                vy = group["Nb"]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(vx, vy, edgecolors=edge_color,
                                marker=list(group[self.marker_column])[0], label=label, facecolors=sample_color,
                                s=self.markersize, zorder=zorder)

        self.plot_arrows()
        self.plot_legend()


class DiagramPearceTY(PearceBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title="Pearce : Ta vs Yb",
                 xlim=(0.1, 100), ylim=(0.05, 50),
                 **kwargs
                 ):
        PearceBase.__init__(self, datasource=datasource, title=title, **kwargs)
        self.xlim = xlim
        self.ylim = ylim
        self.xlabel = "Yb (ppm)"
        self.ylabel = "Ta (ppm)"

    def set_decorations(self):
        PearceBase.set_decorations(self)
        _ = get_translator(self.lang_cfg)

        lx1 = np.array([0.55, 3])
        ly1 = np.array([20, 2])

        lx2 = np.array([0.1, 3])
        ly2 = np.array([0.35, 2])

        lx3 = np.array([3, 5])
        ly3 = np.array([2, 1])

        lx4 = np.array([5, 5])
        ly4 = np.array([0.05, 1])

        lx5 = np.array([5, 100])
        ly5 = np.array([1, 7])

        lx6 = np.array([3, 100])
        ly6 = np.array([2, 20])

        self.ax.plot(lx1, ly1, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx2, ly2, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx3, ly3, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx5, ly5, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx4, ly4, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx6, ly6, color=self.decor_line_col, linewidth=0.6, linestyle='--')

        self.ax.text(0.16, 2.97, _("Syn-COLG"))
        self.ax.text(4.8, 11.3, _("WPG"), )
        self.ax.text(1.03, 0.08, _("VAG"), )
        self.ax.text(16.2, 0.19, _("ORG"), )

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decorations()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.exclude_groups and name not in self.exclude_groups:

                vx = group["Yb"]
                vy = group["Ta"]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(vx, vy, edgecolors=edge_color,
                                marker=list(group[self.marker_column])[0], label=label, facecolors=sample_color,
                                s=self.markersize,
                                zorder=zorder)

        self.plot_legend()



from matplotlib.colors import to_rgba

from georunes.plot.helpers import ArrowDrawerTernary
from georunes.plot.ternary.ternbase import DiagramTernaryBase
from georunes.tools.language import get_translator
from georunes.tools.plotting import normalize_marker_size


# Streckeisen, A., 1974. Classification and nomenclature of plutonic rocks recommendations of the IUGS subcommission
# on the systematics of igneous rocks. Geologische Rundschau, 63(2), pp.773-786.
# Streckeisen, A., 1978. IUGS Subcommission on the Systematics of Igneous Rocks. Classification and Nomenclature of
# Volcanic Rocks, Lamprophyres, Carbonatites and Melilite Rocks. Recommendations and Suggestions. Neues Jahrbuch fur
# Mineralogie. Stuttgart. Abhandlungen, 143, pp.1-14.

class DiagramQAP(DiagramTernaryBase, ArrowDrawerTernary):
    def __init__(self, datasource, title="QAP ternary diagram", decor_set="Plut", decor_text_col="#777777", **kwargs):
        if decor_set not in ("Plut", "Volc"):
            raise ValueError("Parameter 'decor_set' must be 'Plut' or 'Volc'")

        DiagramTernaryBase.__init__(self, datasource=datasource,
                                    top_var='Qnorm', left_var='Anorm', right_var='Pnorm',
                                    title=title,
                                    top_label='Q', left_label='A', right_label='P',
                                    decor_text_col=decor_text_col,
                                    tscale='linear', left_scale='linear', right_scale='linear',
                                    **kwargs
                                    )

        self.decor_set = decor_set

    def check_parameters(self):
        chklist = ('Qnorm', 'Pnorm', 'Anorm')
        columns = self.data.columns
        missing = []
        for e in chklist:
            if e not in columns:
                missing.append(e)

        if len(missing) > 0:
            raise Exception("Parameters " + str(missing) + " are missing for QAP plot.")

    def set_decoration(self, ):
        _ = get_translator(self.lang_cfg)
        # Order for points : right var, top var, left var
        p01 = (80, 20, 0)
        p02 = (0, 20, 80)

        p21 = (40, 60, 0)
        p22 = (0, 60, 40)

        p31 = (10, 90, 0)
        p32 = (0, 90, 10)

        p41 = (40 * 0.1, 60, 40 * 0.9)
        p42 = (10, 0, 90)

        p51 = (80 * 0.35, 20, 80 * 0.65)
        p52 = (35, 00, 65)

        p61 = (40 * 0.65, 60, 40 * 0.35)
        p62 = (65, 00, 35)

        p621 = (40 * 0.65, 60, 40 * 0.35)
        p622 = (80 * 0.65, 20, 80 * 0.35)

        if self.decor_set == "Plut":
            p11 = (95, 5, 0)
            p12 = (0, 5, 95)

            q51 = (60 * 0.35, 40, 60 * 0.65)
            q52 = (80 * 0.35, 20, 80 * 0.65)

            p71 = (40 * 0.9, 60, 40 * 0.1)
            p72 = (90, 00, 10)

        elif self.decor_set == "Volc":
            p11 = (0, 5, 95)
            p12 = (65 * 0.95, 5, 35 * 0.95)

            q51 = (40 * 0.35, 60, 40 * 0.65)
            q52 = (80 * 0.35, 20, 80 * 0.65)

            p71 = (80 * 0.9, 20, 80 * 0.1)
            p72 = (40 * 0.9, 60, 40 * 0.1)

        self.tax.line(p01, p02, linewidth=1, color=self.decor_line_col)
        self.tax.line(p11, p12, linewidth=1, color=self.decor_line_col)
        self.tax.line(p21, p22, linewidth=1, color=self.decor_line_col)
        self.tax.line(p31, p32, linewidth=1, color=self.decor_line_col)
        self.tax.line(p41, p42, linewidth=1, color=self.decor_line_col)
        self.tax.line(p51, p52, linewidth=1, color=self.decor_line_col)
        self.tax.line(q51, q52, linewidth=1, linestyle="--", color=self.decor_line_col)
        self.tax.line(p61, p62, linewidth=1, color=self.decor_line_col)

        if self.decor_set == "Plut":
            self.tax.line(p621, p622, linewidth=1, color=self.decor_line_col)
            self.tax.line(p71, p72, linewidth=1, color=self.decor_line_col)

            # right, top
            self.tax.annotate(_("granite"), (12, 50), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("syeno-\ngranite"), (12, 25), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("monzo-\ngranite"), (31, 25), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("quartz-\nsyenite"), (21, 10), fontsize="small", ha="center", color=self.decor_text_col)
            self.tax.annotate(_("quartz-\nmonzonite"), (45, 10), fontsize="small", ha="center",
                              color=self.decor_text_col)
            self.tax.annotate(_("quartz-\nmonzodiorite"), (69, 10), fontsize="small", ha="center",
                              color=self.decor_text_col)
            self.tax.annotate(_("syenite"), (19, 1), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("monzonite"), (41, 1), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("granodiorite"), (41, 37), rotation=-70, fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("tonalite"), (45, 45), rotation=-65, fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("monzodiorite\nmonzogabbro"), (74, -14, 16), fontsize="small",
                              color=self.decor_text_col)

        if self.decor_set == "Volc":
            self.tax.line(p621, p622, linewidth=1, linestyle="--", color=self.decor_line_col)
            self.tax.line(p71, p72, linewidth=1, linestyle="--", color=self.decor_line_col)
            # right, top
            self.tax.annotate(_("rhyolite"), (14, 42), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("quartz-\ntrachyte"), (23, 10), fontsize="small", ha='center',
                              color=self.decor_text_col)
            self.tax.annotate(_("quartz-\nlatite"), (42, 10), fontsize="small", ha='center', color=self.decor_text_col)
            self.tax.annotate(_("dacite"), (40, 42), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("basalt\nandesite"), (73, 10), fontsize="small", ha='center', color=self.decor_text_col)
            self.tax.annotate(_("trachyte"), (18, 1), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("latite"), (43, 1), fontsize="small", color=self.decor_text_col)
            self.tax.annotate(_("monzodiorite\nmonzogabbro"), (74, -14, 16), fontsize="small",
                              color=self.decor_text_col)

    def plot(self):
        DiagramTernaryBase.plot_config(self, no_gridline=True, no_ticks=self.no_ticks)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):
                var_sum = group[self.top_var] + group[self.left_var] + group[self.right_var]
                norm_z = 100 * group[self.left_var] / var_sum
                norm_x = 100 * group[self.right_var] / var_sum
                norm_y = 100 * group[self.top_var] / var_sum

                if self.marker != '':
                    marker = self.marker
                else:
                    marker = list(group[self.marker_column])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize

                # Convert points to tuples
                points = [(norm_x.get(i), norm_y.get(i), norm_z.get(i)) for i, sample in
                          group[self.top_var].items()]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.tax.scatter(points, edgecolors=edge_color,
                                 marker=marker, label=label, facecolors=sample_color, s=size,
                                 zorder=zorder)

                if self.annotation:
                    for i, sample in group[self.top_var].items():
                        self.tax.annotate(group[self.annotation].get(i), (norm_x.get(i), norm_y.get(i)),
                                          fontsize='xx-small')

        self.plot_legend()



import numpy as np
from matplotlib.colors import to_rgba

from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer, ArrowDrawer
from georunes.tools.language import format_chemical_formula as _fml, get_translator


# Cox, K.G., Bell, J.D. and Pankhurst, R.J., 1979. Petrographic aspects of plutonic rocks. In The Interpretation of
# Igneous Rocks (pp. 283-307). Springer, Dordrecht.
# Le Maitre, R.W., Streckeisen, A., Zanettin, B., Le Bas, M.J., Bonin, B. and Bateman, P., 2005. Igneous Rocks: A
# Classification and Glossary of Terms. Igneous Rocks: A Classification and Glossary of Terms, p.252.

class DiagramSiAlkali(DiagramBase, ArrowDrawer, LegendDrawer):

    def __init__(self, datasource, decor_set="Cox", title="Total alkali versus silica diagram",
                 padding={"bottom": 0.20}, annotation=None, classi_line=True,
                 alpha_color=0.4, alpha_edge_color=0.8,
                 **kwargs):

        if decor_set not in ("Cox", "TAS"):
            raise ValueError("Parameter 'decor_set' must be 'TAS' or 'Cox'")

        DiagramBase.__init__(self, datasource, title=title, padding=padding, **kwargs)

        self.classi_line = classi_line
        self.annotation = annotation
        self.xlabel = _fml("SiO2 (wt%)")
        self.ylabel = _fml("Na2O + K2O (wt%)")
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color
        self.decor_set = decor_set

    def set_decorations(self):
        _ = get_translator(self.lang_cfg)
        self.ax.set_xlim(35, 80)
        self.ax.set_ylim(1, 17)

        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        if self.decor_set == "Cox":
            lx1 = np.array([41, 44, 52, 52, 53, 46.4, 44.5, 41])
            ly1 = np.array([3, 2, 1.75, 5.6, 7.2, 7, 5.75, 3])

            lx2 = np.array([52, 55, 54.75, 52, 52])
            ly2 = np.array([1.75, 1.75, 5.5, 5.6, 1.75])

            lx3 = np.array([55, 57, 63, 62.5, 54.75, 55])
            ly3 = np.array([1.75, 2, 3.5, 7, 5.5, 1.75])
            lx4 = np.array([63, 70, 65, 62.5, 63])
            ly4 = np.array([3.5, 5.5, 9, 7, 3.55])
            lx5 = np.array([70, 75, 75, 69.3, 65, 70])
            ly5 = np.array([5.5, 8, 9, 11.9, 9, 5.5])
            lx6 = np.array([52, 54.75, 62.5, 65, 62, 57, 53, 52])
            ly6 = np.array([5.6, 5.5, 7, 9, 10, 9, 7.2, 5.6])
            lx7 = np.array([39.5, 41, 44.5, 46.4, 48, 50.25, 54.5, 51.5, 49, 40.5, 43.5, 39.5])
            ly7 = np.array([4, 3, 5.75, 7, 8.3, 9.25, 11, 13.2, 15, 9.5, 8.33, 4])
            lx8 = np.array([46.4, 53, 57, 54.25, 50.25, 48, 46.4])
            ly8 = np.array([7, 7.2, 9, 9.35, 9.25, 8.3, 7])
            lx9 = np.array([50.25, 54.25, 57, 62, 65, 69.3, 62, 57.75, 54.5, 50.25])
            ly9 = np.array([9.25, 9.35, 9, 10, 9, 11.9, 14, 11.25, 11, 9.25])
            lx10 = np.array([36, 39.5, 43.5, 40.5, 36, 36])
            ly10 = np.array([5.9, 4, 8.33, 9.5, 6.5, 5.9])
            lx11 = np.array([49, 51.5, 54.5, 57.75, 62, 52.2, 51.5, 49])
            ly11 = np.array([15, 13.2, 11, 11.25, 14, 16.25, 16.25, 15])
            lx12 = np.array([57.75, 62])
            ly12 = np.array([11.25, 10])
            lx13 = np.array([43.5, 45.35, 51.5])
            ly13 = np.array([8.33, 9.375, 13.2])
            lx14 = np.array([45.35, 48])
            ly14 = np.array([9.375, 8.3])
            lx15 = np.array([44.5, 52])
            ly15 = np.array([5.75, 5.6])

            self.ax.plot(lx1, ly1, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx2, ly2, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx3, ly3, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx4, ly4, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx5, ly5, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx6, ly6, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx7, ly7, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx8, ly8, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx9, ly9, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx10, ly10, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx11, ly11, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx12, ly12, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx13, ly13, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx14, ly14, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(lx15, ly15, color=self.decor_line_col, linewidth=0.6)

            if self.classi_line:
                lx17 = np.array([44, 52, 65, 80])
                ly17 = np.array([2, 5.6, 8, 8.35])

                p = np.poly1d(np.polyfit(lx17, ly17, 3))
                t17 = np.arange(44, 76, 1)
                u17 = []
                for val in t17:
                    u17.append(p(val))
                self.ax.plot(t17, u17, color=self.decor_line_col, linewidth=1)

            labels = [
                [68, 9.5, _("Granite")],
                [64, 4.9, _("Quartz diorite")],
                [57, 4.5, _("Diorite")],
                [43, 3.5, _("Gabbro")],
                [59.5, 11.5, _("Syenite")],
                [49, 8, _("Syenodiorite")],
                [50.5, 14, _("Nepheline\nsyenite")],
                [46, 6.2, _("Gabbro")],
                [37, 6.5, _("Ijolite")],
                [56.5, 10, _("Syenite")]
            ]
            for elt in labels:
                self.ax.text(*elt, fontsize="x-small", color=self.decor_text_col)

            self.ax.text(70, 14, _("Alkaline"), color=self.decor_text_col)
            self.ax.text(70, 3, _("Subalkaline"), color=self.decor_text_col)

        elif self.decor_set == 'TAS':
            lx0 = np.array([41, 41, 45])
            ly0 = np.array([3, 7, 9.4])
            hx1 = np.array([45, 48.4, 52.5])
            hy1 = np.array([9.4, 11.5, 14])
            hx2 = np.array([45, 45, 49.4, 53, 57.6])
            hy2 = np.array([1, 5, 7.3, 9.3, 11.7])
            hx3 = np.array([45, 52, 57, 63, 69])
            hy3 = np.array([5, 5, 5.9, 7, 8])

            vx4 = np.array([41, 41, 45])
            vy4 = np.array([1, 3, 3])
            vx5 = np.array([52, 52, 49.4, 45])
            vy5 = np.array([1, 5, 7.3, 9.4])
            vx6 = np.array([57, 57, 53, 48.4])
            vy6 = np.array([1, 5.9, 9.3, 11.5])
            vx7 = np.array([63, 63, 57.6, 52.5, 50.28])
            vy7 = np.array([1, 7, 11.7, 14, 15])
            vx8 = np.array([74, 69, 69])
            vy8 = np.array([1, 8, 13])

            self.ax.plot(lx0, ly0, color=self.decor_line_col, linewidth=0.6, linestyle="--", )
            self.ax.plot(hx1, hy1, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(hx2, hy2, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(hx3, hy3, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(vx4, vy4, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(vx5, vy5, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(vx6, vy6, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(vx7, vy7, color=self.decor_line_col, linewidth=0.6)
            self.ax.plot(vx8, vy8, color=self.decor_line_col, linewidth=0.6)

            labels = [
                [43, 2, _("picro-\nbasalt")],
                [48, 3, _("basalt")],
                [54.5, 3.5, _("basaltic\nandesite")],
                [60, 5, _("andesite")],
                [66.5, 4.5, _("dacite")],
                [49, 5.5, _("trachy-\nbasalt")],
                [53, 6.5, _("basaltic\ntrachy-\nandesite")],
                [57.5, 8.2, _("trachy-\nandesite")],
                [64, 10.5, _("trachyte/\ntrachydacite")],
                [45, 7, _("tephrite/\nbasanite")],
                [49, 9.3, _("phonotephrite")],
                [53, 11.5, _("tephriphonolite")],
                [57, 14, _("phonolite")],
                [40, 9.5, _("foidite")],
            ]
            for elt in labels:
                self.ax.text(*elt, fontsize="x-small", ha="center", color=self.decor_text_col)

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decorations()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.exclude_groups and name not in self.exclude_groups:
                nak = group["Na2O"] + group["K2O"]

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(group["SiO2"], nak, edgecolors=edge_color,
                                marker=list(group[self.marker_column])[0], label=label, facecolors=sample_color,
                                s=self.markersize,
                                zorder=zorder)

                if self.annotation:
                    for i, sample in group["SiO2"].items():
                        self.ax.annotate(group[self.annotation].get(i), (group["SiO2"].get(i), nak[i]),
                                         fontsize='xx-small')

        self.plot_arrows()
        self.plot_legend()



import warnings
import pandas as pd
from georunes.modmin.optim.bvls import BVLS
from georunes.modmin.optim.gd import GradientDescent
from georunes.modmin.optim.nnls import NNLS
from georunes.modmin.optim.randsearch import RandomSearch
from georunes.tools.data import linspace_to_end, round_floor_n_digits
from georunes.tools.filemanager import FileManager
from georunes.tools.warnings import FunctionParameterWarning

filemanager = FileManager.get_instance()


def create_cumulate_finder_from_file(source, sheet_name, *args, **kwargs):
    raw_minerals_data = filemanager.read_file(source, sheet_name)
    return CumulateFinder(raw_minerals_data, *args, **kwargs)


class CumulateFinder:
    def __init__(self, raw_minerals_data, ignore_oxels=None, optimizer='BVLS', norm='euclidian', nb_results=5, verbose=0):
        self.ignore_oxels = ignore_oxels
        self.prepare_data(raw_minerals_data)
        self.nb_results = nb_results
        if optimizer in ('BVLS', 'NNLS') and norm != 'euclidian' :
            warnings.warn("BVLS and NNLS are specialized for euclidian norm. Distance function set to euclidian norm.", FunctionParameterWarning)
            norm = 'euclidian'
        self.optimizer = optimizer
        if self.optimizer == "BVLS":
            self.opt = BVLS(verbose=verbose)
        elif self.optimizer == "NNLS":
            self.opt = NNLS(verbose=verbose)
        elif self.optimizer == "RS":
            self.opt = RandomSearch(verbose=verbose, dist_func=norm)
        elif self.optimizer == "GD":
            self.opt = GradientDescent(verbose=verbose, dist_func=norm, filling_tolerance=0.01)

    def filter_oxels(self, comp, verbose=1):
        removed = []
        # list_keys = list()
        for elox in comp.keys():
            if elox not in self.oxel_list:
                del comp[elox]

        if len(removed) and verbose:
            print("Removed the oxels " + str(removed) + "from a composition.")

    def prepare_data(self, raw_minerals_data):
        if self.ignore_oxels:
            for ox in raw_minerals_data.columns.tolist()[1:]:
                if ox in self.ignore_oxels:
                    raw_minerals_data = raw_minerals_data.drop(columns=ox)
        raw_minerals_data = raw_minerals_data.fillna(0)
        raw_minerals_data = raw_minerals_data[(raw_minerals_data.iloc[:, 1:] != 0).any(axis=1)]
        self.raw_minerals_data = raw_minerals_data  # For optimization, index not needed
        raw_minerals_data = raw_minerals_data.set_index(raw_minerals_data.keys()[0])

        self.oxel_list = raw_minerals_data.columns.tolist()
        raw_minerals_data = raw_minerals_data[[*list(self.oxel_list)]]  # Order oxides as in source
        self.minerals_data = raw_minerals_data.transpose()
        self.list_minerals = self.minerals_data.keys().tolist()
        self.nb_minerals = len(self.list_minerals)

    def compute_beta_limit_removal(self, parent, child):
        div = parent / child
        self.max_beta = div.min()

    def find_cumulate_to_remove(self, parent_comp, child_comp, verbose=0):
        self.filter_oxels(parent_comp)
        self.filter_oxels(child_comp)
        data = pd.DataFrame(index=self.oxel_list, )
        self.compute_beta_limit_removal(parent_comp, child_comp)
        fracts_list = linspace_to_end(round_floor_n_digits(self.max_beta, 3), self.nb_results)[::-1]
        for beta in linspace_to_end(round_floor_n_digits(self.max_beta, 3), self.nb_results)[::-1]:
            x_df = self.get_cumulate_comp(parent_comp, child_comp, beta)
            data['alpha_' + str(round(1 - beta, 3))] = x_df

        data = data.T
        data = data.reset_index()
        data.rename(columns={'index': 'Sample'}, inplace=True)
        data['Total'] = data.iloc[:, 1:].sum(axis=1)

        p,s = None,None
        if self.optimizer in ("BVLS", "NNLS"):
            p, s = self.opt.compute(data, skip_cols=1, raw_minerals_data=self.raw_minerals_data)
        elif self.optimizer == 'RS':
            p, s = self.opt.compute(data, skip_cols=1, raw_minerals_data=self.raw_minerals_data,
                                    max_iter=100000, search_semiedge=0.2, scale_semiedge=0.75, force_totals=False,
                                    unfillable_partitions_allowed=True)  #
        elif self.optimizer == 'GD':
            self.opt.set_verbose(2)
            p, s = self.opt.compute(data, skip_cols=1, raw_minerals_data=self.raw_minerals_data,
                                    max_iter=10000, learn_rate=0.0001, force_totals=False)
        proportions, supplements = p, s

        return proportions, supplements, fracts_list

    @staticmethod
    def get_cumulate_comp(parent, child, beta):
        x_df = (parent - beta * child) / (1 - beta)
        return x_df



from matplotlib.colors import to_rgba

from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer, ArrowDrawer
from georunes.tools.chemistry import molar_ratio
from georunes.tools.language import format_chemical_formula as _fml, get_translator


# Shand, S.J., 1943. Eruptive rocks: their genesis, composition, and classification, with a chapter on meteorites. J.
# Wiley & sons, Incorporated.

class DiagramShand(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title="A/NK vs A/CNK diagram of Shand (1943)",
                 annotation=None,
                 xlim=None, ylim=None,
                 alpha_color=0.4, alpha_edge_color=0.8,
                 **kwargs):

        DiagramBase.__init__(self, datasource=datasource, title=title, **kwargs)

        self.annotation = annotation
        if xlim:
            self.xlim = xlim
        else:
            self.xlim = (0.6, 1.4)
        if ylim:
            self.ylim = ylim
        else:
            self.ylim = (0.4, 2)

        self.xlabel = _fml("Al2O3/(NaO + K2O + CaO)")
        self.ylabel = _fml("Al2O3/(NaO + K2O)")
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color

    def set_decoration(self):
        _ = get_translator(self.lang_cfg)
        self.ax.set_xlim(*self.xlim)
        self.ax.set_ylim(*self.ylim)

        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        x1 = (1, 1)
        y1 = (-2, 2)

        x2 = (1.1, 1.1)
        y2 = (-2, 2)

        x3 = (-2, 2)
        y3 = (-2, 2)

        x4 = (-2, 2)
        y4 = (1, 1)

        self.ax.plot(x1, y1, self.decor_line_col, linewidth=0.8, zorder=1)
        self.ax.plot(x2, y2, self.decor_line_col, linewidth=0.4, zorder=1)
        self.ax.plot(x3, y3, self.decor_line_col, linewidth=0.8, zorder=1)
        self.ax.plot(x4, y4, self.decor_line_col, linewidth=0.8, zorder=1)

        self.ax.text(self.xlim[0] + 0.02, 0.85, _("Peralkaline"))
        self.ax.text(self.xlim[0] + 0.02, 1.85, _("Metaluminous"))
        self.ax.text(1.15, 1.85, _("Peraluminous"))

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):
                acnk = molar_ratio(group["Al2O3"]) / (
                        molar_ratio(group["Na2O"]) + molar_ratio(group["K2O"]) + molar_ratio(group["CaO"]))
                ank = molar_ratio(group["Al2O3"]) / (molar_ratio(group["Na2O"]) + molar_ratio(group["K2O"]))

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(acnk, ank, edgecolors=edge_color,
                                marker=list(group[self.marker_column])[0], label=label, facecolors=sample_color,
                                s=self.markersize,
                                zorder=zorder)

                if self.annotation:
                    for i, sample in acnk.items():
                        self.ax.annotate(group[self.annotation].get(i), (acnk.get(i), ank.get(i)), fontsize='xx-small')

        self.plot_arrows()
        self.plot_legend()



import warnings
import pandas as pd
from georunes.tools.preprocessing import check_geochem_res

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

eml = pkg_resources.files('georunes.data').joinpath('reservoirs.csv')


# Singleton model, from github.com/pazdera/1098129
class Reservoirs:
    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if Reservoirs.__instance is None:
            Reservoirs()
        return Reservoirs.__instance

    def __init__(self, source=eml):
        """ Virtually private constructor. """
        if Reservoirs.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            Reservoirs.__instance = self

        data = pd.read_csv(source, )
        data.set_index("alias")
        self.model_list = list(data.loc[:, "alias"])
        self.compos = {}
        for model in self.model_list:
            self.compos[model] = data.loc[data['alias'] == model].squeeze()  # Convert to a Series

    def get_models_list(self):
        return self.model_list if self.model_list else None

    def get_label(self, model):
        return self.compos[model]["reservoir"]

    def get_color(self, model):
        return self.compos[model]["color"]

    def get_chondrite_compo(self, element):
        return self.compos['CI'][element]

    def register_model(self, model):
        alias = model['alias']
        self.model_list.append(alias)

        d = {}
        for el, val in model.items():
            if el != 'alias':
                d[el] = val

        self.compos[alias] = pd.Series(d)

        return self.compos[alias]


def chondrite_val(element):
    res = Reservoirs.get_instance()
    return res.get_chondrite_compo(element)


def get_reservoir_norm(norm_slug, default="CI"):
    res = Reservoirs.get_instance()
    if isinstance(norm_slug, dict) and check_geochem_res(norm_slug):
        norm = res.register_model(norm_slug)
    elif norm_slug in res.model_list:
        norm = res.compos[norm_slug]
    else:
        if norm_slug is not None:
            msg = "The data of reservoir " + norm_slug + " are not found. Check the figure configuration."
            warnings.warn(msg)
        norm = res.compos[default]
    return norm



from matplotlib.colors import to_rgba

from georunes.plot.helpers import ArrowDrawerTernary
from georunes.plot.ternary.ternbase import DiagramTernaryBase
from georunes.tools.language import get_translator
from georunes.tools.plotting import normalize_marker_size


# Irvine, T.N. and Baragar, W.R.A., 1971. A Guide to the Chemical Classification of the Common Volcanic Rocks.
# Canadian Journal of Earth Sciences, 8(5), pp.523-548.

class DiagramAFM(DiagramTernaryBase, ArrowDrawerTernary):
    def __init__(self, datasource, title="AFM diagram (Irvine and Baragar 1971)", **kwargs):
        DiagramTernaryBase.__init__(self, datasource=datasource,
                                    title=title,
                                    top_label='F', left_label='A', right_label='M',
                                    tscale='linear', left_scale='linear', right_scale='linear',
                                    legend_ncol=4,
                                    **kwargs)

    def set_decoration(self, ):
        _ = get_translator(self.lang_cfg)
        # Order for points : right var, top var, left var

        p11 = (4, 36, 60)
        p12 = (15.5, 53, 31.5)
        p13 = (19, 55, 26)
        p14 = (23, 55, 28)
        p15 = (27.5, 53, 19.5)
        p16 = (50, 35, 15)

        limit = [p11, p12, p13, p14, p15, p16]
        self.tax.plot(limit, linewidth=1.0, color=self.decor_line_col)
        self.tax.annotate(_("Tholeiite series"), (5, 70,), fontsize="small")
        self.tax.annotate(_("Calc-alkaline series"), (30, 10,), fontsize="small")

    def plot(self):
        DiagramTernaryBase.plot_config(self, no_ticks=True)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):
                top_var = group['Fe2O3t']
                left_var = group['Na2O'] + group['K2O']
                right_var = group['MgO']

                var_sum = top_var + left_var + right_var
                norm_z = 100 * left_var / var_sum
                norm_x = 100 * right_var / var_sum
                norm_y = 100 * top_var / var_sum

                if self.marker != '':
                    marker = self.marker
                else:
                    marker = list(group[self.marker_column])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize

                points = []
                for i, sample in top_var.items():
                    pos = (norm_x.get(i), norm_y.get(i), norm_z.get(i))
                    points.append(pos)

                label = list(group[self.label_column])[0] if self.label_defined else name
                zorder = 4
                if self.zorder_column:
                    zorder = list(group[self.zorder_column])[0]

                sample_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group[self.color_column])[0], alpha=self.alpha_edge_color)
                self.tax.scatter(points, edgecolors=edge_color,
                                 marker=marker, label=label, facecolors=sample_color, s=size,
                                 zorder=zorder)

                if self.annotation:
                    for i, sample in top_var.items():
                        self.tax.annotate(group[self.annotation].get(i), (norm_x.get(i), norm_y.get(i)),
                                          fontsize='xx-small')

        self.plot_legend()
